import pandas as pd
import torch
import openai
import sys
from transformers import AutoTokenizer, AutoModelForCausalLM
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_faithfulness_scoring 
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_relevance_scoring
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_context_relevance_scoring
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_context_relevance_scoring_togetherai
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_faithfulness_scoring_togetherai
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_relevance_scoring_togetherai
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_context_relevance_scoring_claude 
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_faithfulness_scoring_claude
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_relevance_scoring_claude
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_context_relevance_scoring_vllm
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_faithfulness_scoring_vllm
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_relevance_scoring_vllm

if 'ipykernel' in sys.modules:
    # We are in a Jupyter notebook or similar (uses IPython kernel)
    from tqdm.notebook import tqdm
else:
    # We are in a regular Python environment (e.g., terminal or script)
    from tqdm import tqdm

def ues_idp_config(in_domain_prompts_dataset: str, unlabeled_evaluation_set: str, context_relevance_system_prompt: str, answer_relevance_system_prompt: str, answer_faithfulness_system_prompt: str, debug_mode: bool, documents: int, model_choice: str, vllm: bool, host_url: str, request_delay: int) -> dict:
    """Configures UES and IDP for evaluation"""

    # Check if both vllm and host_url are provided together
    if vllm and host_url == "None":
        raise ValueError("host_url must be provided if vllm is True.")
    elif not vllm and host_url != "None": 
        raise ValueError("vLLM must be set to True if host_url is provided")
    
    if in_domain_prompts_dataset is not None:
            in_domain_prompts_dataset = pd.read_csv(in_domain_prompts_dataset, sep='\t')
    else:
        in_domain_prompts_dataset = None 

    if unlabeled_evaluation_set is not None:
        unlabeled_evaluation_set = pd.read_csv(unlabeled_evaluation_set, sep='\t')
    else:
        unlabeled_evaluation_set = None
    if in_domain_prompts_dataset is None and unlabeled_evaluation_set is None: 
        print(f"Error: UES and IDP are not provided")
        exit()
    if in_domain_prompts_dataset is None: 
        print(f"Error: IDP is not provided")
        exit()
    if unlabeled_evaluation_set is None: 
        print(f"Error: UES is not provided")
        exit()

    if documents > len(unlabeled_evaluation_set): 
        sys.exit("Error: documents size passed in is larger than documents present in unlabeled evaluation set")
    
    if documents == 0: 
        documents = len(unlabeled_evaluation_set)

    # context_relevance_answers = unlabeled_evaluation_set["Context_Relevance_Label"].tolist()
    # answer_relevance_answers = unlabeled_evaluation_set["Answer_Relevance_Label"].tolist()
    # answer_faithfulness_answers = unlabeled_evaluation_set["Answer_Faithfulness_Label"].tolist()
    context_relevance_scores = []
    answer_relevance_scores = []
    answer_faithfulness_scores = []

    required_columns = ['Query', 'Document', 'Answer']
    missing_columns = [col for col in required_columns if col not in unlabeled_evaluation_set.columns]
    if missing_columns:
        print(f"Missing columns in the DataFrame: {missing_columns}")


    if "gpt" in model_choice:
        with tqdm(total=documents, desc=f"Evaluating large subset with {model_choice}") as pbar:
            for index, row in unlabeled_evaluation_set[:documents].iterrows():
                # Extract query, document, and answer from the row
                try:
                    query = row['Query']
                except KeyError:
                    query = row['Question']
                
                try:
                    _ = in_domain_prompts_dataset.iloc[0]['Query']
                    query_id = "Query"
                except KeyError:
                    try:
                        _ = in_domain_prompts_dataset.iloc[0]['Question']
                        query_id = "Question"
                    except KeyError:
                        sys.exit("Both 'Query' and 'Question' keys are missing for the given row.")

                document = row['Document']
                answer = row['Answer']
                                
                # Scoring
                if vllm: 
                    context_score = few_shot_context_relevance_scoring_vllm(
                    context_relevance_system_prompt, query, document, model_choice, query_id, debug_mode, host_url, request_delay, in_domain_prompts_dataset) 
                    
                    # If context relevance is 0, set answer scores to 0.
                    if context_score == 0:
                        answer_relevance_score = 0
                        answer_faithfulness_score = 0
                    else:
                        answer_relevance_score = few_shot_answer_relevance_scoring_vllm(
                        answer_relevance_system_prompt, query, document, answer, model_choice, query_id, debug_mode, host_url, request_delay, in_domain_prompts_dataset)

                        answer_faithfulness_score = few_shot_answer_faithfulness_scoring_vllm(answer_faithfulness_system_prompt, query, document, answer, model_choice, query_id, debug_mode, host_url, request_delay, in_domain_prompts_dataset)

                else: 
                    context_score = few_shot_context_relevance_scoring(
                        context_relevance_system_prompt, query, document, model_choice, query_id, debug_mode, request_delay, in_domain_prompts_dataset)
                    
                    # If context relevance is 0, set answer scores to 0.
                    if context_score == 0:
                        answer_relevance_score = 0
                        answer_faithfulness_score = 0
                    else:
                        answer_relevance_score = few_shot_answer_relevance_scoring(
                            answer_relevance_system_prompt, query, document, answer, model_choice, query_id, debug_mode, request_delay, in_domain_prompts_dataset)
                        
                        answer_faithfulness_score = few_shot_answer_faithfulness_scoring(answer_faithfulness_system_prompt, query, document, answer, model_choice, query_id, debug_mode, request_delay, in_domain_prompts_dataset)

                # Append scores to respective lists
                context_relevance_scores.append(context_score)
                answer_relevance_scores.append(answer_relevance_score)
                answer_faithfulness_scores.append(answer_faithfulness_score)

                pbar.update(1)
    elif "claude" in model_choice:
        with tqdm(total=documents, desc=f"Evaluating large subset with {model_choice}") as pbar:
            for index, row in unlabeled_evaluation_set[:documents].iterrows():
                # Extract query, document, and answer from the row
                try:
                    query = row['Query']
                except KeyError:
                    query = row['Question']
                
                try:
                    _ = in_domain_prompts_dataset.iloc[0]['Query']
                    query_id = "Query"
                except KeyError:
                    try:
                        _ = in_domain_prompts_dataset.iloc[0]['Question']
                        query_id = "Question"
                    except KeyError:
                        sys.exit("Both 'Query' and 'Question' keys are missing for the given row.")

                document = row['Document']
                answer = row['Answer']

                # Scoring
                context_score = few_shot_context_relevance_scoring_claude(
                    context_relevance_system_prompt, query, document, model_choice, query_id, debug_mode, request_delay, in_domain_prompts_dataset)

                # If context relevance is 0, set answer scores to 0.
                if context_score == 0:
                    answer_relevance_score = 0
                    answer_faithfulness_score = 0
                else: 
                    answer_relevance_score = few_shot_answer_relevance_scoring_claude(
                        answer_relevance_system_prompt, query, document, answer, model_choice, query_id, debug_mode, request_delay, in_domain_prompts_dataset)
                    
                    answer_faithfulness_score = few_shot_answer_faithfulness_scoring_claude(answer_faithfulness_system_prompt, query, document, answer, model_choice, query_id, debug_mode, request_delay, in_domain_prompts_dataset)

                # Append scores to respective lists
                context_relevance_scores.append(context_score)
                answer_relevance_scores.append(answer_relevance_score)
                answer_faithfulness_scores.append(answer_faithfulness_score)

                pbar.update(1)
    else: 
        with tqdm(total=documents, desc=f"Evaluating large subset with {model_choice}") as pbar:
            for index, row in unlabeled_evaluation_set[:documents].iterrows():
                # Extract query, document, and answer from the row
                try:
                    query = row['Query']
                except KeyError:
                    query = row['Question']
                
                try:
                    _ = in_domain_prompts_dataset.iloc[0]['Query']
                    query_id = "Query"
                except KeyError:
                    try:
                        _ = in_domain_prompts_dataset.iloc[0]['Question']
                        query_id = "Question"
                    except KeyError:
                        sys.exit("Both 'Query' and 'Question' keys are missing for the given row.")

                document = row['Document']
                answer = row['Answer']
                
                # Scoring
                if vllm: 
                    context_score = few_shot_context_relevance_scoring_vllm(
                    context_relevance_system_prompt, query, document, model_choice, query_id, debug_mode, host_url, request_delay, in_domain_prompts_dataset) 
                    
                    # If context relevance is 0, set answer scores to 0.
                    if context_score == 0:
                        answer_relevance_score = 0
                        answer_faithfulness_score = 0
                    else: 
                        answer_relevance_score = few_shot_answer_relevance_scoring_vllm(
                        answer_relevance_system_prompt, query, document, answer, model_choice, query_id, debug_mode, host_url, request_delay, in_domain_prompts_dataset)

                        answer_faithfulness_score = few_shot_answer_faithfulness_scoring_vllm(answer_faithfulness_system_prompt, query, document, answer, model_choice, query_id, debug_mode, host_url, request_delay, in_domain_prompts_dataset)
                    
                else: 
                    context_score = few_shot_context_relevance_scoring_togetherai(
                        context_relevance_system_prompt, query, document, model_choice, query_id, debug_mode, request_delay, in_domain_prompts_dataset)
                    
                    # If context relevance is 0, set answer scores to 0.
                    if context_score == 0:
                        answer_relevance_score = 0
                        answer_faithfulness_score = 0
                    else:
                        answer_relevance_score = few_shot_answer_relevance_scoring_togetherai(
                            answer_relevance_system_prompt, query, document, answer, model_choice, query_id, debug_mode, request_delay, in_domain_prompts_dataset)
                        
                        answer_faithfulness_score = few_shot_answer_faithfulness_scoring_togetherai(answer_faithfulness_system_prompt, query, document, answer, model_choice, query_id, debug_mode, request_delay, in_domain_prompts_dataset)

                # Append scores to respective lists
                context_relevance_scores.append(context_score)
                answer_relevance_scores.append(answer_relevance_score)
                answer_faithfulness_scores.append(answer_faithfulness_score)

                pbar.update(1)

    # Convert score lists to DataFrame
    results_df = pd.DataFrame({
        'Context_Relevance_Score': context_relevance_scores,
        'Answer_Relevance_Score': answer_relevance_scores,
        'Answer_Faithfulness_Score': answer_faithfulness_scores
    })

    # Apply filtering to remove invalid rows
    valid_results_df = results_df[
        ~((results_df['Context_Relevance_Score'] == 0) & 
        ((results_df['Answer_Relevance_Score'] == 1) | 
        (results_df['Answer_Faithfulness_Score'] == 1)))
    ]

    print("Original rows:", len(results_df))
    print("Rows after filtering:", len(valid_results_df))

    # Return the average scores from the filtered DataFrame
    return {
        "Context Relevance Scores": round(valid_results_df['Context_Relevance_Score'].mean(), 3),
        "Answer Faithfulness Scores": round(valid_results_df['Answer_Faithfulness_Score'].mean(), 3),
        "Answer Relevance Scores": round(valid_results_df['Answer_Relevance_Score'].mean(), 3)
    }

    


from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import begin
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import filter_dataset
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import preprocess_data
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import load_model
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import evaluate_model
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import post_process_predictions
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import evaluate_and_scoring_data

def rag_scoring_config(alpha, num_trials, evaluation_datasets, few_shot_examples_filepath, checkpoints, labels,
GPT_scoring, annotated_datapoints_filepath, model_choice, llm_judge, assigned_batch_size, number_of_labels, gold_label_path, rag_type, vllm, host_url, request_delay, debug_mode):
    # Validate inputs and determine model loading strategy
    if checkpoints:
        if llm_judge and llm_judge != "None":
            print("Warning: Both checkpoint and llm_judge were provided. Using checkpoints.")
        model_loader = lambda chk: load_model(model_choice, number_of_labels, chk)
    elif llm_judge and llm_judge != "None":
        model_loader = lambda _: load_model(llm_judge, number_of_labels, vllm, None)
    else:
        raise ValueError("No valid model or checkpoint provided.")

    # Use zip only if checkpoints are not empty, otherwise assume only llm_judge is used
    if checkpoints:
        # Here we assume that the length of checkpoints and labels is the same
        pairings = zip(checkpoints, labels)
    else:
        # If no checkpoints, create dummy pairs for labels with None for checkpoint
        pairings = ((None, label) for label in labels)

    for checkpoint, label_column in pairings:
        LLM_judge_ratio_predictions = []
        validation_set_lengths = []
        validation_set_ratios = []
        ppi_confidence_intervals = []
        accuracy_scores = []
        for test_set_selection in evaluation_datasets:

            few_shot_examples = begin(evaluation_datasets, checkpoints, labels, GPT_scoring, few_shot_examples_filepath)

            context_relevance_system_prompt, answer_faithfulness_system_prompt, answer_relevance_system_prompt = filter_dataset(rag_type)
            
            test_set, text_column = preprocess_data(test_set_selection, label_column, labels)

            model, tokenizer, device = model_loader(checkpoint)

            total_predictions, total_references, results, metric = evaluate_model(test_set, label_column, text_column, device, checkpoint, tokenizer, model, assigned_batch_size, model_choice, context_relevance_system_prompt, answer_faithfulness_system_prompt, answer_relevance_system_prompt, few_shot_examples_filepath, llm_judge, vllm, host_url, request_delay, debug_mode)

            test_set, Y_labeled_dataset, Y_labeled_dataloader, Y_labeled_predictions, Yhat_unlabeled_dataset, prediction_column = post_process_predictions(checkpoint, test_set, label_column, total_predictions, labels, gold_label_path, tokenizer, assigned_batch_size, device) 

            evaluate_and_scoring_data(test_set, Y_labeled_predictions, Y_labeled_dataset, Y_labeled_dataloader, Yhat_unlabeled_dataset, alpha, num_trials, model, device, model_choice, 
            annotated_datapoints_filepath, context_relevance_system_prompt, answer_faithfulness_system_prompt, answer_relevance_system_prompt,
            few_shot_examples, metric, prediction_column, label_column, test_set_selection, 
            LLM_judge_ratio_predictions, validation_set_lengths, validation_set_ratios, 
            ppi_confidence_intervals, accuracy_scores, results, checkpoint, llm_judge, vllm, host_url, request_delay, debug_mode)

# if not checkpoints and not llm_judge:
#         raise ValueError("Either checkpoints or an llm_model must be provided.")

#     if checkpoints:
#         if len(checkpoints) != len(evaluation_datasets):
#             raise ValueError("The number of checkpoints must match the number of evaluation datasets.")
#         models_to_use = checkpoints
#     else:
#         models_to_use = [llm_judge] * len(evaluation_datasets)

#     for model_to_use, test_set_selection in zip(models_to_use, evaluation_datasets):
#         for label_column in labels:
#             LLM_judge_ratio_predictions = []
#             validation_set_lengths = []
#             validation_set_ratios = []
#             ppi_confidence_intervals = []
#             accuracy_scores = []

#             few_shot_examples = begin(evaluation_datasets, model_to_use, labels, GPT_scoring, few_shot_examples_filepath)

#             context_relevance_system_prompt, answer_faithfulness_system_prompt, answer_relevance_system_prompt = filter_dataset(rag_type)
            
#             test_set, text_column = preprocess_data(test_set_selection, label_column, labels)

#             model, tokenizer, device = load_model(llm_judge, number_of_labels, GPT_scoring, model_to_use) 

#             total_predictions, total_references, results, metric = evaluate_model(test_set, label_column, text_column, device, GPT_scoring, tokenizer, model, assigned_batch_size, llm_judge, context_relevance_system_prompt, answer_faithfulness_system_prompt, answer_relevance_system_prompt)

#             test_set, Y_labeled_dataset, Y_labeled_dataloader, Y_labeled_predictions, Yhat_unlabeled_dataset, prediction_column = post_process_predictions(test_set, label_column, total_predictions, labels, use_pseudo_human_labels, gold_label_path, tokenizer, assigned_batch_size, device) 

#             evaluate_and_scoring_data(test_set, Y_labeled_predictions, Y_labeled_dataset, Y_labeled_dataloader, Yhat_unlabeled_dataset, alpha, num_trials, model, device, llm_judge, 
#             annotated_datapoints_filepath, context_relevance_system_prompt, answer_faithfulness_system_prompt, answer_relevance_system_prompt,
#             few_shot_examples, metric, prediction_column, label_column, test_set_selection, 
#             LLM_judge_ratio_predictions, validation_set_lengths, validation_set_ratios, 
#             ppi_confidence_intervals, accuracy_scores, results)



from cProfile import label
import torch.nn as nn
from transformers import T5Tokenizer, T5EncoderModel, T5ForConditionalGeneration
from transformers import BertModel, AutoTokenizer, AutoModel, GPT2Tokenizer
#import tensorflow as tf
import sys

import pandas as pd
import numpy as np
import ast
import datasets
from datasets import load_metric
from transformers import TrainingArguments, Trainer

import pyarrow as pa
import pyarrow.dataset as ds

from torch.optim import Adam
from torch.utils.data import DataLoader
from transformers import get_scheduler, AutoModelForCausalLM, AutoConfig, AutoModelForSequenceClassification #MptForSequenceClassification

import torch
from tqdm.auto import tqdm
import statistics
import time

import subprocess as sp
import os
from sklearn.model_selection import train_test_split
import json
import random
import re
import scipy.stats as stats
import argparse

import subprocess
import json

import openai

from tqdm import tqdm
tqdm.pandas()


from ares.RAG_Automatic_Evaluation.ppi import clt_iid, binomial_iid, pp_mean_iid_asymptotic
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import calculate_accuracy, few_shot_context_relevance_scoring
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_faithfulness_scoring, few_shot_answer_relevance_scoring
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_context_relevance_scoring_togetherai, few_shot_answer_faithfulness_scoring_togetherai, few_shot_answer_relevance_scoring_togetherai
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_context_relevance_scoring_claude, few_shot_answer_faithfulness_scoring_claude, few_shot_answer_relevance_scoring_claude
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_context_relevance_scoring_vllm
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_relevance_scoring_vllm
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_faithfulness_scoring_vllm
#############################################################

random_state = 44

np.random.seed(random_state)
random.seed(random_state)
torch.manual_seed(random_state)
os.environ['PYTHONHASHSEED'] = str(random_state)
os.environ["HUGGINGFACE_HUB_DISABLE_DOWNLOAD_PROGRESS"] = "1"
os.environ["TRANSFORMERS_VERBOSITY"] = "error"

############################################################

class CustomBERTModel(nn.Module):
    def __init__(self, number_of_labels, model_choice):
          self.model_choice = model_choice

          super(CustomBERTModel, self).__init__()
          if model_choice in ["mosaicml/mpt-7b-instruct", "mosaicml/mpt-7b"]:

            config = AutoConfig.from_pretrained(model_choice, trust_remote_code=True)
            config.attn_config['attn_impl'] = 'triton'  # change this to use triton-based FlashAttention
            config.max_seq_len = max_token_length

            model_encoding = AutoModelForCausalLM.from_pretrained(
                model_choice,
                config=config,
                #max_seq_len=max_token_length,
                torch_dtype=torch.bfloat16,
                trust_remote_code=True,
                use_auth_token=True
            )
            embedding_size = 4096
            self.encoderModel = model_encoding.transformer

          elif model_choice in ['mosaicml/mpt-1b-redpajama-200b']:

            model_encoding = MptForSequenceClassification.from_pretrained("mosaicml/mpt-1b-redpajama-200b", trust_remote_code=True)
            embedding_size = 2048
            self.encoderModel = model_encoding
          
          elif model_choice in ["google/t5-large-lm-adapt", "google/t5-xl-lm-adapt"]:

            model_encoding = AutoModelForSequenceClassification.from_pretrained(model_choice)
            embedding_size = 1024
            self.encoderModel = model_encoding#.transformer

          elif model_choice in ["roberta-large", "microsoft/deberta-v3-large"]:

            model_encoding = AutoModel.from_pretrained(model_choice)
            embedding_size = 1024
            self.encoderModel = model_encoding

          elif model_choice in ["microsoft/deberta-v2-xlarge", "microsoft/deberta-v2-xxlarge"]:

            model_encoding = AutoModel.from_pretrained(model_choice)
            embedding_size = 1536
            self.encoderModel = model_encoding
          
          else:

            model_encoding = AutoModel.from_pretrained(model_choice)
            embedding_size = 768
            self.encoderModel = model_encoding

          #########################################
          
          self.encoderModel.eval()
          self.classifier = nn.Sequential(nn.Linear(embedding_size, 256), nn.Linear(256, number_of_labels))
          self.embedding_size = embedding_size


    def forward(self, ids, mask, labels=None, decoder_input_ids=None):
        model_choice = self.model_choice
        if model_choice in ["t5-small", "google/t5-xl-lm-adapt", "google/t5-large-lm-adapt", "mosaicml/mpt-1b-redpajama-200b"]:
            total_output = self.encoderModel(input_ids=ids, attention_mask=mask) #labels=labels
            return total_output['logits']
        else:
            total_output = self.encoderModel(ids, attention_mask=mask)
            sequence_output = total_output['last_hidden_state']

            last_hidden_state_formatted = sequence_output[:,0,:].view(-1, self.embedding_size)
            linear2_output = self.classifier(last_hidden_state_formatted)

            return linear2_output

############################################################

def combine_query_document(query: str, document: str, answer=None):
    cleaned_document = re.sub(r'\n+', '\n', document.replace("\r"," ").replace("\t"," ")).strip()
    cleaned_document = cleaned_document.replace("=", " ").replace("-", " ")
    cleaned_document = re.sub(r'\s+', ' ', cleaned_document).strip()
    cleaned_document = (" ").join(cleaned_document.split(" ")[:512])

    if len(query.split(" ")) > 100:
        query = (" ").join(query.split(" ")[:30])

    if answer is None:
        return query + " | " + cleaned_document
    else:
        try:
            return query + " | " + cleaned_document + " | " + answer
        except:
            print("Error with combine_query_document")
            print("Query: " + str(query))
            print("Cleaned Document: " + str(cleaned_document))
            print("Answer: " + str(answer))
            #return str(query) + " | " + str(cleaned_document) + " | " + str(answer)
            return "Error"

def tokenize_function(tokenizer, examples):

    return tokenizer(examples["text"], padding="max_length", truncation=True)#.input_ids

############################################################

def prepare_dataset_for_evaluation(dataframe, label_column: str, text_column: str, assigned_batch_size, tokenizer):
    from datasets.utils.logging import disable_progress_bar
    disable_progress_bar()

    test_set_text = [dataframe.iloc[i][text_column] for i in range(len(dataframe))]
    test_set_label = [dataframe.iloc[i][label_column] for i in range(len(dataframe))]

    test_dataset_pandas = pd.DataFrame({'label': test_set_label, 'text': test_set_text})
    test_dataset_arrow = pa.Table.from_pandas(test_dataset_pandas)
    test_dataset_arrow = datasets.Dataset(test_dataset_arrow)

    classification_dataset = datasets.DatasetDict({'test' : test_dataset_arrow})
    tokenized_datasets = classification_dataset.map(lambda examples: tokenize_function(tokenizer, examples), batched=True)
    tokenized_datasets = tokenized_datasets.remove_columns(["text"])
    tokenized_datasets = tokenized_datasets.rename_column("label", "labels")
    tokenized_datasets.set_format("torch")

    eval_dataloader = DataLoader(tokenized_datasets['test'], batch_size=assigned_batch_size)
    return eval_dataloader

############################################################

def calculate_ppi(Y_labeled,  Yhat_labeled, Yhat_unlabeled, alpha, num_trials):

    n_max = Y_labeled.shape[0]
    # ns = np.linspace(100,n_max,20).astype(int)
    ns = np.linspace(0, n_max, 20).astype(int)

    # Imputed-only estimate
    imputed_estimate = (Yhat_labeled.sum() + Yhat_unlabeled.sum()) / (Yhat_labeled.shape[0] + Yhat_unlabeled.shape[0])

    # Run prediction-powered inference and classical inference for many values of n
    ci = np.zeros((num_trials, ns.shape[0], 2))
    ci_classical = np.zeros((num_trials, ns.shape[0], 2))

    for j in tqdm(range(num_trials), desc="Trials"):  # Wrap the outer loop with tqdm for the progress bar
        for i, n in enumerate(ns):  # Iterate over ns with an index
            rand_idx = np.random.permutation(Y_labeled.shape[0])
            f = Yhat_labeled.astype(float)[rand_idx[:n]]
            y = Y_labeled.astype(float)[rand_idx[:n]]
            output = pp_mean_iid_asymptotic(y, f, Yhat_unlabeled, alpha)
            ci[j, i, :] = output
            # Classical interval
            try:
                ci_classical[j, i, :] = binomial_iid(n, alpha, y.mean())
            except:
                avg_ci_classical = None

    avg_ci = ci.mean(axis=0)[-1]

    try:
        ci_imputed = binomial_iid(Yhat_unlabeled.shape[0], alpha, imputed_estimate)
    except:
        ci_imputed = None
    try:
        avg_ci_classical = ci_classical.mean(axis=0)[-1]
    except:
        avg_ci_classical = None

    return avg_ci, avg_ci_classical, ci_imputed

######################################################################

def begin(evaluation_datasets, checkpoints, labels, GPT_scoring, few_shot_examples_filepath):
    print("--------------------------------------------------------")
    print("Evaluation Sets: " + str(evaluation_datasets))
    print("Checkpoints: " + str(checkpoints))
    print("Labels: "  + str(labels))
    print("--------------------------------------------------------")

    ######################################################################

    # if GPT_scoring:
    #     checkpoint = ["" for _ in range(len(labels))]

    if few_shot_examples_filepath is not None:
        few_shot_examples = pd.read_csv(few_shot_examples_filepath, sep="\t")
        # print("few_shot_examples")
        # print(len(few_shot_examples))
        # print(few_shot_examples.head())
    return few_shot_examples

####################################################################

def clean_document(document: str):
    cleaned_document = re.sub(r'\n+', '\n', document.replace("\r"," ").replace("\t"," ")).strip()
    cleaned_document = cleaned_document.replace("=", " ").replace("-", " ")
    cleaned_document = re.sub(r'\s+', ' ', cleaned_document).strip()
    cleaned_document = (" ").join(cleaned_document.split(" ")) #[:512]
    return cleaned_document

def clean_query(query: str):
    cleaned_query = query.replace("\n", " ").replace("\r"," ").replace("\t"," ").strip()
    return cleaned_query

#################################################
def filter_dataset(rag_type: str = "question_answering"):
    if rag_type == "question_answering": 
        context_relevance_system_prompt = "Given the following question and document, you must analyze the provided document and determine whether it is sufficient for answering the question. "
        context_relevance_system_prompt += "In your evaluation, you should consider the content of the document and how it relates to the provided question. "
        context_relevance_system_prompt += 'Output your final verdict by strictly following this format: "[[Yes]]" if the document is sufficient and "[[No]]" if the document provided is not sufficient. '
        context_relevance_system_prompt += "Do not provide any additional explanation for your decision.\n\n"
    
        answer_faithfulness_system_prompt = "Given the following question, document, and answer, you must analyze the provided answer and determine whether it is faithful to the contents of the document. "
        answer_faithfulness_system_prompt += "The answer must not offer new information beyond the context provided in the document. "
        answer_faithfulness_system_prompt += "The answer also must not contradict information provided in the document. "
        answer_faithfulness_system_prompt += 'Output your final verdict by strictly following this format: "[[Yes]]" if the answer is faithful to the document and "[[No]]" if the answer is not faithful to the document. '
        answer_faithfulness_system_prompt += "Do not provide any additional explanation for your decision.\n\n"

        answer_relevance_system_prompt = "Given the following question, document, and answer, you must analyze the provided answer and document before determining whether the answer is relevant for the provided question. "
        answer_relevance_system_prompt += "In your evaluation, you should consider whether the answer addresses all aspects of the question and provides only correct information from the document for answering the question. "
        answer_relevance_system_prompt += 'Output your final verdict by strictly following this format: "[[Yes]]" if the answer is relevant for the given question and "[[No]]" if the answer is not relevant for the given question. '
        answer_relevance_system_prompt += "Do not provide any additional explanation for your decision.\n\n"

    elif rag_type == "fact_checking": 
        context_relevance_system_prompt = "You are an expert fact-checking agent. "
        context_relevance_system_prompt += "Given the following statement and document, you must analyze the provided document and determine whether it is sufficient for determining the statement's factuality. "
        context_relevance_system_prompt += "In your evaluation, you should consider the content of the document and how it relates to the provided statement's factuality. "
        context_relevance_system_prompt += 'Output your final verdict by strictly following this format: "[[Yes]]" if the document is sufficient and "[[No]]" if the document is not sufficient. '
        context_relevance_system_prompt += "Do not provide any additional explanation for your decision.\n\n"

        answer_faithfulness_system_prompt = "Given the following statement, document, and answer, you must analyze the provided answer and determine whether it is faithful to the contents of the document. "

        answer_relevance_system_prompt = "Given the following statement, document, and answer, you must analyze the provided answer and document before determining whether the answer is relevant for the provided statement. "
        answer_relevance_system_prompt += "In your evaluation, you should consider whether the answer addresses all aspects of the statement and provides only correct information from the document for answering the statement. "
        answer_relevance_system_prompt += 'Output your final verdict by strictly following this format: "[[Yes]]" if the answer is relevant for the given statement and "[[No]]" if the answer is not relevant for the given statement. '
        answer_relevance_system_prompt += "Do not provide any additional explanation for your decision.\n\n"

    elif rag_type == "dialogue_agent":
        context_relevance_system_prompt = "You are an expert dialogue agent. "
        context_relevance_system_prompt += "Given the following dialogue and document, you must analyze the provided document and determine whether it is relevant for responding to the dialogue. "
        context_relevance_system_prompt += "In your evaluation, you should consider the content of the document and how it relates to the provided dialogue. "
        context_relevance_system_prompt += 'Output your final verdict by strictly following this format: "[[Yes]]" if the document is relevant and "[[No]]" if the document provided is not relevant. '
        context_relevance_system_prompt += "Do not provide any additional explanation for your decision.\n\n"

        answer_faithfulness_system_prompt = "Given the following dialogue, document, and answer, you must analyze the provided answer and determine whether it is faithful to the contents of the document. "
        
        answer_relevance_system_prompt = "Given the following dialogue, document, and answer, you must analyze the provided answer and document before determining whether the answer is relevant for the provided dialogue. "
        answer_relevance_system_prompt += "In your evaluation, you should consider whether the answer addresses all aspects of the dialogue and provides only correct information from the document for responding to the dialogue. "
        answer_relevance_system_prompt += 'Output your final verdict by strictly following this format: "[[Yes]]" if the answer is relevant for the given dialogue and "[[No]]" if the answer is not relevant for the given dialogue. '
        answer_relevance_system_prompt += "Do not provide any additional explanation for your decision.\n\n"
    
    else: 
        sys.exit("Error: rag_type not specified, please specifiy in configuration with 'rag_type'")
    
    return context_relevance_system_prompt, answer_faithfulness_system_prompt, answer_relevance_system_prompt

def preprocess_data(test_set_selection, label_column, labels): 
    test_set = pd.read_csv(test_set_selection, sep="\t")
    text_column = 'concat_text'
    test_set = test_set[test_set[label_column].notna()]
    if "Context" in label_column:
        test_set[text_column] = [combine_query_document(test_set.iloc[i]['Query'], test_set.iloc[i]['Document']) for i in range(len(test_set))]
    else:
        test_set[text_column] = [combine_query_document(test_set.iloc[i]['Query'], test_set.iloc[i]['Document'], test_set.iloc[i]['Answer']) for i in range(len(test_set))]

    test_set = test_set[test_set[text_column] != "Error"]
    # print("Example Text for " + label_column + " Scoring")
    # print(test_set.iloc[10][text_column])
    if len(test_set) < 10:
        # print("Example Text for " + label_column + " Scoring")
        # print(test_set.iloc[10][text_column])
        raise ValueError("Insufficient Data: Dataset has fewer than 10 rows after filtering!")
    
    return test_set, text_column

        ############################################################

def togetherai_list_models(api_key):
    if not api_key:
        return []
    try:
        # Running the command to list models
        result = subprocess.run(['together', 'models', 'list'], capture_output=True, text=True, check=True)
        lines = result.stdout.split('\n')
        models = []
        for line in lines[3:]:
            if '|' in line:
                parts = line.split('|')
                if len(parts) > 1:
                    model_name = parts[1].strip()  
                    models.append(model_name)
        return models
    except subprocess.CalledProcessError as e:
        print(f"An error occurred: {e.stderr}")
        return ["N/A"]  
    except Exception as e:
        print(f"An unexpected error occurred: {e}")
        return ["N/A"]

def load_model(model_identifier, number_of_labels, vllm, checkpoint=None):
    check = False
    tokenizer = None
    api_key = os.getenv("TOGETHER_API_KEY")
    
    together_models = togetherai_list_models(api_key)
    if "gpt" in model_identifier:
        check = True
    elif "claude" in model_identifier: 
        check = True
    elif model_identifier in together_models: 
        check = True 
    elif vllm: 
        check = True
    else: 
        check = False
    
    if check == False:
        max_token_length = 2048
        tokenizer = AutoTokenizer.from_pretrained(model_identifier, model_max_length=max_token_length)

    # Prepare device setup
    torch.cuda.empty_cache()
    device = torch.device("cuda:0")

    if check is False: 
        # Initialize model
        model = CustomBERTModel(number_of_labels, model_identifier)
        model.to(device)
    else: 
        model = model_identifier

    # Load the model from a checkpoint if provided
    if checkpoint:
        checkpoint_dict = torch.load(checkpoint)
        if "encoderModel.embeddings.position_ids" in checkpoint_dict:
            del checkpoint_dict["encoderModel.embeddings.position_ids"]
        model.load_state_dict(checkpoint_dict)
        print("Loaded model from checkpoint:", checkpoint)
    else:
        print("Loaded model based on model identifier:", model_identifier)
    
    return model, tokenizer, device

    ############################################################

def evaluate_model(test_set, label_column, text_column, device, checkpoint, tokenizer, model, assigned_batch_size, model_choice, 
context_relevance_system_prompt, answer_faithfulness_system_prompt, answer_relevance_system_prompt, few_shot_examples_filepath, llm_judge, vllm, host_url, request_delay, debug_mode):
     
    metric = load_metric("accuracy")

    if checkpoint:
        total_predictions = torch.FloatTensor([]).to(device)
        total_references = torch.FloatTensor([]).to(device)
        total_logits = torch.FloatTensor([]).to(device)
        eval_dataloader = prepare_dataset_for_evaluation(test_set, label_column, text_column, assigned_batch_size, tokenizer)
        model.eval()
        with tqdm(eval_dataloader, desc="Evaluating", leave=False) as progress_bar:
            for batch in progress_bar:

                with torch.no_grad():

                    if model_choice in ["mosaicml/mpt-1b-redpajama-200b"]:
                        new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].bool().to(device)}
                    else:
                        new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].to(device)}

                    if model_choice in ["t5-small", "google/t5-xl-lm-adapt", "google/t5-large-lm-adapt"]:
                        new_batch['decoder_input_ids'] = batch['labels'].reshape(batch['labels'].shape[0], 1).to(device)

                    outputs = model(**new_batch)

                    logits = outputs
                    predictions = torch.argmax(logits, dim=-1)
                    metric.add_batch(predictions=predictions, references=batch['labels'].to(device))

                    total_predictions = torch.cat((total_predictions, predictions), 0)
                    total_references = torch.cat((total_references, batch['labels'].to(device)), 0)
                    total_logits = torch.cat((total_logits, logits), 0)

                    progress_bar.update(1)

    else:
        print("Performing Model scoring!")
        # print("Using gpt model: " + gpt_model)
        # if perform_zero_shot:
        #     # print("Using zero-shot approach")
        #     # print("Setting few-shot example to None...")
        #     few_shot_examples = None

        few_shot_examples = pd.read_csv(few_shot_examples_filepath, sep="\t")
        # Few Shot Dataset Edge Check: Query ID 
        try:
                _ = few_shot_examples.iloc[0]['Query']
                query_id = "Query"
        except KeyError:
            try:
                _ = few_shot_examples.iloc[0]['Question']
                query_id = "Question"
            except KeyError:
                sys.exit("Both 'Query' and 'Question' keys are missing for the given row in few shot dataset.")
        # # Labeled Dataset Edge Check: Query ID 
        # try:
        #         _ = Y_labeled_dataset.iloc[0]['Query']
        #         query_labeled_id = "Query"
        # except KeyError:
        #     try:
        #         _ = Y_labeled_dataset.iloc[0]['Question']
        #         query_labeled_id = "Question"
        #     except KeyError:
        #         sys.exit("Both 'Query' and 'Question' keys are missing in labeled dataset.")

        if "Context_Relevance_Label" == label_column:
            if vllm: 
                test_set["Context_Relevance_Prediction"] = test_set.progress_apply(lambda x: few_shot_context_relevance_scoring_vllm(context_relevance_system_prompt, clean_query(x[query_id]), x["Document"], llm_judge, query_id, debug_mode, host_url, request_delay, few_shot_examples), axis=1)
            # tqdm.pandas(desc="Generating context relevance scores...", total=test_set.shape[0])
            test_set["Context_Relevance_Prediction"] = test_set.progress_apply(lambda x: few_shot_context_relevance_scoring(context_relevance_system_prompt, clean_query(x[query_id]), x["Document"], llm_judge, query_id, debug_mode, few_shot_examples), axis=1)
        elif "Answer_Faithfulness_Label" == label_column:
            if vllm: 
                test_set["Answer_Faithfulness_Prediction"] = test_set.progress_apply(lambda x: few_shot_answer_faithfulness_scoring_vllm(answer_faithfulness_system_prompt, clean_query(x[query_id]), x["Document"], x["Answer"], llm_judge, query_id, debug_mode, host_url, request_delay, few_shot_examples), axis=1)
            # tqdm.pandas(desc="Generating answer faithfulness scores...", total=test_set.shape[0])
            test_set["Answer_Faithfulness_Prediction"] = test_set.progress_apply(lambda x: few_shot_answer_faithfulness_scoring(answer_faithfulness_system_prompt, clean_query(x[query_id]), x["Document"], x["Answer"], llm_judge, query_id, debug_mode, few_shot_examples), axis=1)
        if "Answer_Relevance_Label" == label_column:
            if vllm: 
                test_set["Answer_Relevance_Prediction"] = test_set.progress_apply(lambda x: few_shot_answer_relevance_scoring_vllm(answer_relevance_system_prompt, clean_query(x[query_id]), x["Document"], x["Answer"], llm_judge, query_id, debug_mode, host_url, request_delay, few_shot_examples), axis=1)
            # tqdm.pandas(desc="Generating answer relevance scores...", total=test_set.shape[0])
            test_set["Answer_Relevance_Prediction"] = test_set.progress_apply(lambda x: few_shot_answer_relevance_scoring(answer_relevance_system_prompt, clean_query(x[query_id]), x["Document"], x["Answer"], llm_judge, query_id, debug_mode, few_shot_examples), axis=1)

        total_predictions = test_set[label_column.replace("_Label", "_Prediction")].to_numpy()
        total_references = test_set[label_column].to_numpy()
     
    results = metric.compute(references=total_references, predictions=total_predictions)

    return total_predictions, total_references, results, metric

        ########################################################################
def post_process_predictions(checkpoint, test_set, label_column, total_predictions, labels, gold_label_path, tokenizer, assigned_batch_size, device): 
    prediction_column = label_column + "_Model_Predictions"
    test_set[prediction_column] = total_predictions.tolist()
    test_set = test_set[test_set[label_column].notna()]
    for label in labels:
        if label != label_column:
            test_set = test_set[test_set[label] != 0]

        # print("Gathering ML predictions for Y_labeled_dataset in PPI!")

        Y_labeled_dataset = pd.read_csv(gold_label_path, sep="\t")
        Y_labeled_dataset = Y_labeled_dataset[Y_labeled_dataset[label_column].notna()]
        Y_labeled_dataset = Y_labeled_dataset.head(300)

        text_column = 'concat_text'
        if "Context" in label_column:
            Y_labeled_dataset[text_column] = [combine_query_document(Y_labeled_dataset.iloc[i]['Query'], Y_labeled_dataset.iloc[i]['Document']) for i in range(len(Y_labeled_dataset))]
        else:
            Y_labeled_dataset[text_column] = [combine_query_document(Y_labeled_dataset.iloc[i]['Query'], Y_labeled_dataset.iloc[i]['Document'], Y_labeled_dataset.iloc[i]['Answer']) for i in range(len(Y_labeled_dataset))]

        Y_labeled_dataset = Y_labeled_dataset[Y_labeled_dataset[text_column] != "Error"]
        
        if checkpoint:
            Y_labeled_dataloader = prepare_dataset_for_evaluation(Y_labeled_dataset, label_column, text_column, assigned_batch_size, tokenizer)
        else: 
            Y_labeled_dataloader = None
            
        Y_labeled_predictions = torch.FloatTensor([]).to(device)

        Yhat_unlabeled_dataset = test_set
        ####################################

    return test_set, Y_labeled_dataset, Y_labeled_dataloader, Y_labeled_predictions, Yhat_unlabeled_dataset, prediction_column

def evaluate_and_scoring_data(test_set, Y_labeled_predictions, Y_labeled_dataset, Y_labeled_dataloader, Yhat_unlabeled_dataset, 
alpha, num_trials, model, device, model_choice, swap_human_labels_for_gpt4_labels, context_relevance_system_prompt, answer_faithfulness_system_prompt, answer_relevance_system_prompt, few_shot_examples, metric, prediction_column, 
label_column, test_set_selection, LLM_judge_ratio_predictions, validation_set_lengths, validation_set_ratios, ppi_confidence_intervals, accuracy_scores, results, checkpoint, llm_judge, vllm, host_url, request_delay, debug_mode):
    # progress_bar = tqdm(range(len(Y_labeled_dataloader)))

    if checkpoint:
        model.eval()
        with tqdm(Y_labeled_dataloader, desc="Scoring", leave=False) as progress_bar:
            for batch in progress_bar:

                with torch.no_grad():

                    if model_choice in ["mosaicml/mpt-1b-redpajama-200b"]:
                        new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].bool().to(device)}
                    else:
                        new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].to(device)}

                    if model_choice in ["t5-small", "google/t5-xl-lm-adapt", "google/t5-large-lm-adapt"]:
                        new_batch['decoder_input_ids'] = batch['labels'].reshape(batch['labels'].shape[0], 1).to(device)

                    outputs = model(**new_batch)

                    logits = outputs
                    predictions = torch.argmax(logits, dim=-1)
                    metric.add_batch(predictions=predictions, references=batch['labels'].to(device))

                    Y_labeled_predictions = torch.cat((Y_labeled_predictions, predictions), 0)

                    progress_bar.update(1)

        Y_labeled_dataset[prediction_column] = Y_labeled_predictions.detach().cpu().numpy().tolist()
        
        Yhat_unlabeled_dataset = test_set

    ############################################################

    else:
        if llm_judge == "None":
            sys.exit("Error: No llm_judge provided")
        
        elif "gpt" in llm_judge:
            Y_labeled_predictions = []

            # Few Shot Dataset Edge Check: Query ID 
            try:
                    _ = few_shot_examples.iloc[0]['Query']
                    query_id = "Query"
            except KeyError:
                try:
                    _ = few_shot_examples.iloc[0]['Question']
                    query_id = "Question"
                except KeyError:
                    sys.exit("Both 'Query' and 'Question' keys are missing for the given row in few shot dataset.")

            # Labeled Dataset Edge Check: Query ID 
            try:
                    _ = Y_labeled_dataset.iloc[0]['Query']
                    query_labeled_id = "Query"
            except KeyError:
                try:
                    _ = Y_labeled_dataset.iloc[0]['Question']
                    query_labeled_id = "Question"
                except KeyError:
                    sys.exit("Both 'Query' and 'Question' keys are missing in labeled dataset.")
            with tqdm(total=len(Y_labeled_dataset), desc="Evaluating", leave=False) as progress_bar:
                for _, row in Y_labeled_dataset.iterrows():
                    query = row[query_labeled_id]
                    document = row["Document"]
                    answer = row["Answer"]
                    if "Context_Relevance_Label" == label_column:
                        if vllm: # Check if vLLM host
                            Y_labeled_predictions.append(few_shot_context_relevance_scoring_vllm(context_relevance_system_prompt, query, document, model_choice, query_id, debug_mode, host_url, request_delay, few_shot_examples))
                            progress_bar.update(1)
                        Y_labeled_predictions.append(few_shot_context_relevance_scoring(context_relevance_system_prompt, clean_query(query), document, model, query_id, debug_mode, few_shot_examples))
                        progress_bar.update(1)
                    elif "Answer_Faithfulness_Label" == label_column:
                        if vllm:
                            Y_labeled_predictions.append(few_shot_answer_faithfulness_scoring_vllm(answer_faithfulness_system_prompt, query, document, answer, model_choice, query_id, debug_mode, host_url, request_delay, few_shot_examples))
                            progress_bar.update(1)
                        Y_labeled_predictions.append(few_shot_answer_faithfulness_scoring(answer_faithfulness_system_prompt, clean_query(query), document, answer, model, query_id, debug_mode, few_shot_examples))
                        progress_bar.update(1)
                    elif "Answer_Relevance_Label" == label_column:
                        if vllm:
                            Y_labeled_predictions.append(few_shot_answer_relevance_scoring_vllm(context_relevance_system_prompt, query, document, answer, model_choice, query_id, debug_mode, host_url, request_delay, few_shot_examples))
                            progress_bar.update(1)
                        Y_labeled_predictions.append(few_shot_answer_relevance_scoring(answer_relevance_system_prompt, clean_query(query), document, answer, model, query_id, debug_mode, few_shot_examples))
                        progress_bar.update(1)
            Y_labeled_predictions_np = np.array(Y_labeled_predictions)
            Y_labeled_dataset[prediction_column] = Y_labeled_predictions_np.tolist()
        elif "claude" in llm_judge: 
            Y_labeled_predictions = []

            # Few Shot Dataset Edge Check: Query ID 
            try:
                    _ = few_shot_examples.iloc[0]['Query']
                    query_id = "Query"
            except KeyError:
                try:
                    _ = few_shot_examples.iloc[0]['Question']
                    query_id = "Question"
                except KeyError:
                    sys.exit("Both 'Query' and 'Question' keys are missing for the given row in few shot dataset.")

            # Labeled Dataset Edge Check: Query ID 
            try:
                    _ = Y_labeled_dataset.iloc[0]['Query']
                    query_labeled_id = "Query"
            except KeyError:
                try:
                    _ = Y_labeled_dataset.iloc[0]['Question']
                    query_labeled_id = "Question"
                except KeyError:
                    sys.exit("Both 'Query' and 'Question' keys are missing in labeled dataset.")
            with tqdm(total=len(Y_labeled_dataset), desc="Evaluating", leave=False) as progress_bar:
                for _, row in Y_labeled_dataset.iterrows():
                    query = row[query_labeled_id]
                    document = row["Document"]
                    answer = row["Answer"]
                    if "Context_Relevance_Label" == label_column:
                        Y_labeled_predictions.append(few_shot_context_relevance_scoring_claude(context_relevance_system_prompt, clean_query(query), document, model, query_id, debug_mode, few_shot_examples))
                        progress_bar.update(1)
                    elif "Answer_Faithfulness_Label" == label_column:
                        Y_labeled_predictions.append(few_shot_answer_faithfulness_scoring_claude(answer_faithfulness_system_prompt, clean_query(query), document, answer, model, query_id, debug_mode, few_shot_examples))
                        progress_bar.update(1)
                    elif "Answer_Relevance_Label" == label_column:
                        Y_labeled_predictions.append(few_shot_answer_relevance_scoring_claude(answer_relevance_system_prompt, clean_query(query), document, answer, model, query_id, debug_mode, few_shot_examples))
                        progress_bar.update(1)
            Y_labeled_predictions_np = np.array(Y_labeled_predictions)
            Y_labeled_dataset[prediction_column] = Y_labeled_predictions_np.tolist()
        else: 
            Y_labeled_predictions = []

            # Few Shot Dataset Edge Check: Query ID 
            try:
                    _ = few_shot_examples.iloc[0]['Query']
                    query_id = "Query"
            except KeyError:
                try:
                    _ = few_shot_examples.iloc[0]['Question']
                    query_id = "Question"
                except KeyError:
                    sys.exit("Both 'Query' and 'Question' keys are missing for the given row in few shot dataset.")

            # Labeled Dataset Edge Check: Query ID 
            try:
                    _ = Y_labeled_dataset.iloc[0]['Query']
                    query_labeled_id = "Query"
            except KeyError:
                try:
                    _ = Y_labeled_dataset.iloc[0]['Question']
                    query_labeled_id = "Question"
                except KeyError:
                    sys.exit("Both 'Query' and 'Question' keys are missing in labeled dataset.")
            with tqdm(total=len(Y_labeled_dataset), desc="Evaluating", leave=False) as progress_bar:
                for _, row in Y_labeled_dataset.iterrows():
                    query = row[query_labeled_id]
                    document = row["Document"]
                    answer = row["Answer"]
                    if "Context_Relevance_Label" == label_column:
                        if vllm: # Check if vLLM host
                            Y_labeled_predictions.append(few_shot_context_relevance_scoring_vllm(context_relevance_system_prompt, query, document, model_choice, query_id, debug_mode, host_url, request_delay, few_shot_examples))
                            progress_bar.update(1)
                        Y_labeled_predictions.append(few_shot_context_relevance_scoring_togetherai(context_relevance_system_prompt, clean_query(query), document, model, query_id, debug_mode, few_shot_examples))
                        progress_bar.update(1)
                    elif "Answer_Faithfulness_Label" == label_column:
                        if vllm: # Check if vLLM host
                            Y_labeled_predictions.append(few_shot_context_relevance_scoring_vllm(context_relevance_system_prompt, query, document, model_choice, query_id, debug_mode, host_url, request_delay, few_shot_examples))
                            progress_bar.update(1)
                        Y_labeled_predictions.append(few_shot_answer_faithfulness_scoring_togetherai(answer_faithfulness_system_prompt, clean_query(query), document, answer, model, query_id, debug_mode, few_shot_examples))
                        progress_bar.update(1)
                    elif "Answer_Relevance_Label" == label_column:
                        if vllm: # Check if vLLM host
                            Y_labeled_predictions.append(few_shot_context_relevance_scoring_vllm(context_relevance_system_prompt, query, document, model_choice, query_id, debug_mode, host_url, request_delay, few_shot_examples))
                            progress_bar.update(1)
                        Y_labeled_predictions.append(few_shot_answer_relevance_scoring_togetherai(answer_relevance_system_prompt, clean_query(query), document, answer, model, query_id, debug_mode, few_shot_examples))
                        progress_bar.update(1)
            Y_labeled_predictions_np = np.array(Y_labeled_predictions)
            Y_labeled_dataset[prediction_column] = Y_labeled_predictions_np.tolist()

    ############################################################

    # if swap_human_labels_for_gpt4_labels:
    #     if "Context_Relevance_Label" == label_column:
    #         Y_labeled_dataset[label_column] = Y_labeled_dataset.progress_apply(lambda x: few_shot_context_relevance_scoring(context_relevance_system_prompt, clean_query(x["Query"]), x["Document"], gpt_model, few_shot_examples), axis=1)
    #     elif "Answer_Faithfulness_Label" == label_column:
    #         Y_labeled_dataset[label_column] = Y_labeled_dataset.progress_apply(lambda x: few_shot_answer_faithfulness_scoring(answer_faithfulness_system_prompt, clean_query(x["Query"]), x["Document"], x["Answer"], gpt_model, few_shot_examples), axis=1)
    #     elif "Answer_Relevance_Label" == label_column:
    #         Y_labeled_dataset[label_column] = Y_labeled_dataset.progress_apply(lambda x: few_shot_answer_relevance_scoring(answer_relevance_system_prompt, clean_query(x["Query"]), x["Document"], x["Answer"], gpt_model, few_shot_examples), axis=1)
    #     else:
    #         print("Error! Could not generate GPT labels for PPI.")
    #         assert False 
        
    Y_labeled = Y_labeled_dataset[label_column].values.astype(int)
    Yhat_labeled = Y_labeled_dataset[prediction_column].values.astype(int)
    Yhat_unlabeled = Yhat_unlabeled_dataset[prediction_column].values.astype(int)
    
    # print("Y_labeled, Yhat_labeled, Yhat_unlabeled for " + test_set_selection + " - " + label_column)
    # print(len(Y_labeled))
    # print(len(Yhat_labeled))
    # print(len(Yhat_unlabeled))
    # print("Y_labeled_dataset Label Distribution: ")
    # print(Y_labeled_dataset[label_column].tolist().count(1))
    # print(Y_labeled_dataset[label_column].tolist().count(0))
    # print("Y_labeled_dataset Prediction Distribution: ")
    # print(Y_labeled_dataset[prediction_column].tolist().count(1))
    # print(Y_labeled_dataset[prediction_column].tolist().count(0))
    # print("Yhat_unlabeled_dataset Prediction Distribution: ")
    # print(Yhat_unlabeled_dataset[prediction_column].tolist().count(1))
    # print(Yhat_unlabeled_dataset[prediction_column].tolist().count(0))

    ######################################################################

    avg_ci, avg_ci_classical, ci_imputed = calculate_ppi(Y_labeled,  Yhat_labeled, Yhat_unlabeled, alpha, num_trials)
    LLM_judge_prediction = sum(avg_ci) / len(avg_ci)
    LLM_judge_ratio_predictions.append(LLM_judge_prediction)
    validation_set_lengths.append(len(test_set))
    ppi_confidence_intervals.append([round(value, 3) for value in avg_ci])
    accuracy_scores.append(round(results['accuracy'], 3))

    # Check if the ground truth label column exists and has any non-null values
    if label_column in Yhat_unlabeled_dataset.columns and not Yhat_unlabeled_dataset[label_column].isnull().all():
        # Calculate the ground truth ratio only if the column is valid and not entirely null
        validation_set_ratios.append(round(Yhat_unlabeled_dataset[label_column].tolist().count(1) / len(Yhat_unlabeled_dataset), 3))
        ground_truth_available = True
    else:
        ground_truth_available = False

    ######################################################################

    indexed_list = list(enumerate(LLM_judge_ratio_predictions))
    sorted_list = sorted(indexed_list, key=lambda x: x[1])
    sorted_indices = [index for index, _ in sorted_list]

    print("--------------------------------------------------")
    print(label_column + " Scoring")
    print("ARES Ranking")
    print("ARES Prediction: " + str(LLM_judge_ratio_predictions))
    print("ARES Confidence Interval: " + str(ppi_confidence_intervals))
    print("Number of Examples in Evaluation Set: " + str(validation_set_lengths))
    if ground_truth_available: 
        print("Ground Truth Performance: " + str(validation_set_ratios))
    print("ARES LLM Judge Accuracy on Ground Truth Labels: " + str(accuracy_scores))
    print("Annotated Examples used for PPI: " + str(len(Y_labeled)))
    print("--------------------------------------------------\n")

# if __name__ == '__main__':

#     parser = argparse.ArgumentParser()

#     parser.add_argument("--alpha", type=float, required=True)
#     parser.add_argument("--num_trials", type=int, required=True)
#     parser.add_argument("--evaluation_datasets", nargs='+', required=True)
#     parser.add_argument("--checkpoints", nargs='+', required=True)
#     parser.add_argument("--labels", nargs='+', required=True)

#     parser.add_argument("--GPT_scoring", type=str, default="False", required=True)
#     parser.add_argument("--gpt_model", type=str, default="gpt-3.5-turbo-16k", required=False)
#     parser.add_argument("--perform_zero_shot", type=str, default="False", required=False)
#     parser.add_argument("--few_shot_examples_filepath", type=str, required=True)

#     parser.add_argument("--Y_labeled_count", type=int, default=300, required=False)
#     parser.add_argument("--use_pseudo_human_labels", type=str, default="False", required=False)
#     parser.add_argument("--gold_label_path", type=str, required=False)
#     parser.add_argument("--swap_human_labels_for_gpt_labels", type=str, default="False", required=False)

#     args = parser.parse_args()

#     ### Instructions

#     # Settings for Human-labeled gold set for PPI
#     alpha = args.alpha
#     num_trials = args.num_trials
#     evaluation_datasets = args.evaluation_datasets
#     checkpoints = args.checkpoints
#     labels = args.labels
    
#     # Settings for zero/few-shot GPT scoring
#     GPT_scoring = args.GPT_scoring
#     if GPT_scoring == "True":
#         GPT_scoring = True
#     else:
#         GPT_scoring = False
    
#     gpt_model = args.gpt_model
#     perform_zero_shot = args.perform_zero_shot
#     if perform_zero_shot == "True":
#         perform_zero_shot = True
#     else:
#         perform_zero_shot = False
#     few_shot_examples_filepath = args.few_shot_examples_filepath

#     Y_labeled_count = args.Y_labeled_count
#     use_pseudo_human_labels = args.use_pseudo_human_labels
#     if use_pseudo_human_labels == "True":
#         use_pseudo_human_labels = True
#     else:
#         use_pseudo_human_labels = False
#     gold_label_path = args.gold_label_path
    
#     swap_human_labels_for_gpt4_labels = args.swap_human_labels_for_gpt_labels
#     if swap_human_labels_for_gpt4_labels == "True":
#         swap_human_labels_for_gpt4_labels = True
#     else:
#         swap_human_labels_for_gpt4_labels = False

#     assigned_batch_size = 1
#     number_of_labels = 2

    ############################################################



####################################################################

# for checkpoint, label_column in zip(checkpoints, labels):

#     LLM_judge_ratio_predictions = []
#     validation_set_lengths = []
#     validation_set_ratios = []
#     ppi_confidence_intervals = []
#     accuracy_scores = []
#     for test_set_selection in evaluation_datasets:


from .LLM_as_a_Judge_Adaptation.Generate_Synthetic_Queries_and_Answers import load_model
from .LLM_as_a_Judge_Adaptation.Generate_Synthetic_Queries_and_Answers import load_documents
from .LLM_as_a_Judge_Adaptation.Generate_Synthetic_Queries_and_Answers import load_few_shot_prompt
from .LLM_as_a_Judge_Adaptation.Generate_Synthetic_Queries_and_Answers import generate_contradictory_answers
from .LLM_as_a_Judge_Adaptation.Generate_Synthetic_Queries_and_Answers import generate_few_shot_prompts
from .LLM_as_a_Judge_Adaptation.Generate_Synthetic_Queries_and_Answers import save_synthetic_queries
from .LLM_as_a_Judge_Adaptation.Generate_Synthetic_Queries_and_Answers import Generate_Synthetic_Answers
# from .LLM_as_a_Judge_Adaptation.Generate_Synthetic_Queries_and_Answers import print_synthetic_queries

def synthetic_generator_config(document_filepaths: list, few_shot_prompt_filename: str,
                               synthetic_queries_filenames: list, documents_sampled: int,
                               model_choice: str = "google/flan-t5-xxl", flan_approach: bool = True, clean_documents: bool = False,
                               regenerate_synth_questions: bool = True, 
                               percentiles: list = [0.05, 0.25, 0.5, 0.95], 
                               question_temperatures: list = [2.0, 1.5, 1.0, 0.5, 0.0],
                               regenerate_answers: bool = True,
                               generate_contradictory_answers_with_flan: bool = True, 
                               number_of_negatives_added_ratio: float = 0.5, # Check whether can also be an int
                               lower_bound_for_negatives: int = 5, # Need to be an int value
                               number_of_contradictory_answers_added_ratio: float = 0.67, # Check whether can also be an int
                               number_of_positives_added_ratio: float = 0.0, # Check whether can also be an int
                               regenerate_embeddings: float = True, 
                               synthetic_query_prompt: str = "You are an expert question-answering system. You must create a question for the provided document. The question must be answerable within the context of the document.\n\n"
                               ): 
    
    model, tokenizer, device = load_model(flan_approach, model_choice)

    if len(document_filepaths) != len(synthetic_queries_filenames):
        raise ValueError("document_filepaths and synthetic_queries_filenames lists must be of the same length.")

    for document_filepath, synthetic_queries_filename in zip(document_filepaths, synthetic_queries_filenames):
        for_fever_dataset = False
        if "fever" in document_filepath.lower():
            for_fever_dataset = True
        for_wow_dataset = False
        if "wow" in document_filepath.lower():
            for_wow_dataset = True

        documents = load_documents(document_filepath, clean_documents, documents_sampled)

        few_shot_examples, length_of_fewshot_prompt = load_few_shot_prompt(few_shot_prompt_filename,for_fever_dataset,for_wow_dataset)

        few_shot_examples_for_contradictory_answers = generate_contradictory_answers(few_shot_prompt_filename,generate_contradictory_answers_with_flan,for_fever_dataset,for_wow_dataset)

        answer_gen_few_shot_examples, length_of_fewshot_prompt_answer_gen = generate_few_shot_prompts(few_shot_prompt_filename,for_fever_dataset,for_wow_dataset)

        save_synthetic_queries(documents, regenerate_synth_questions, flan_approach, few_shot_examples, 
        length_of_fewshot_prompt, device, tokenizer, model, percentiles, for_fever_dataset, for_wow_dataset, 
        synthetic_query_prompt, synthetic_queries_filename, question_temperatures)

        Generate_Synthetic_Answers(synthetic_queries_filename,
        regenerate_answers, flan_approach, answer_gen_few_shot_examples, length_of_fewshot_prompt_answer_gen, 
        device, tokenizer, model, for_fever_dataset, for_wow_dataset, generate_contradictory_answers_with_flan,
        few_shot_examples_for_contradictory_answers, number_of_negatives_added_ratio, lower_bound_for_negatives, number_of_contradictory_answers_added_ratio, number_of_positives_added_ratio, regenerate_embeddings)

        # print_synthetic_queries(synthetic_queries_filename)
    


from .synthetic_generator import synthetic_generator_config
from .binary_classifier import binary_classifer_config
from .rag_scoring import rag_scoring_config
from .ues_idp import ues_idp_config
from .kilt_filter import KILT_dataset_process
from .superglue_filter import superGlue
from typing import List
import pandas as pd

context_relevance_system_prompt = "You are an expert dialogue agent."
context_relevance_system_prompt += "Your task is to analyze the provided document and determine whether it is relevant for responding to the dialogue."
context_relevance_system_prompt += "In your evaluation, you should consider the content of the document and how it relates to the provided dialogue."
context_relevance_system_prompt += 'Output your final verdict by strictly following this format: "[[Yes]]" if the document is relevant and "[[No]]" if the document provided is not relevant.'
context_relevance_system_prompt += "Do not provide any additional explanation for your decision.\n\n"
context_relevance_system_prompt = context_relevance_system_prompt

answer_relevance_system_prompt = "Given the following question, document, and answer, you must analyze the provided answer and document before determining whether the answer is relevant for the provided question."
answer_relevance_system_prompt += "In your evaluation, you should consider whether the answer addresses all aspects of the question and provides only correct information from the document for answering the question."
answer_relevance_system_prompt += 'Output your final verdict by strictly following this format: "[[Yes]]" if the answer is relevant for the given question and "[[No]]" if the answer is not relevant for the given question.'
answer_relevance_system_prompt += "Do not provide any additional explanation for your decision.\n\n"
answer_relevance_system_prompt = answer_relevance_system_prompt

answer_faithfulness_system_prompt = "Given the following question, document, and answer, you must analyze the provided answer and determine whether it is faithful to the contents of the document."
answer_faithfulness_system_prompt += "The answer must not offer new information beyond the context provided in the document."
answer_faithfulness_system_prompt += "The answer also must not contradict information provided in the document."
answer_faithfulness_system_prompt += 'Output your final verdict by strictly following this format: "[[Yes]]" if the answer is faithful to the document and "[[No]]" if the answer is not faithful to the document.'
answer_faithfulness_system_prompt += "Do not provide any additional explanation for your decision.\n\n"
answer_faithfulness_system_prompt = answer_faithfulness_system_prompt

class ARES: 
    REQUIRED_BUT_HAS_DEFAULT = object()

    config_spec = {
        "ues_idp": {
            "in_domain_prompts_dataset": (str, None), # Required parameter (No default),
            "unlabeled_evaluation_set" : (str, None), # Required parameter (No default),
            "context_relevance_system_prompt": (str, context_relevance_system_prompt),
            "answer_relevance_system_prompt": (str, answer_relevance_system_prompt),
            "answer_faithfulness_system_prompt": (str, answer_faithfulness_system_prompt),
            "debug_mode": (bool, False),
            "documents": (int, 0),
            "model_choice": (str, "gpt-3.5-turbo-1106"),
            "request_delay": (int, 0),
            "vllm": (bool, False),
            "host_url": (str, "None")
        },

        "synthetic_query_generator": {
            "document_filepaths": (list, None),  # Required parameter (No default)
            "few_shot_prompt_filename": (str, None),  # Required parameter (No default)
            "synthetic_queries_filenames": (list, None),  # Required parameter (No default)
            "documents_sampled": (int, None),  # Required parameter (No default)
            "model_choice": (str, "google/flan-t5-xxl"),
            "flan_approach": (bool, True), 
            "clean_documents": (bool, False), 
            "regenerate_synth_questions": (bool, True), 
            "percentiles": (list,[0.05, 0.25, 0.5, 0.95]), 
            "question_temperatures": (list, [2.0, 1.5, 1.0, 0.5, 0.0]), 
            "regenerate_answers": (bool, True), 
            "generate_contradictory_answers_with_flan": (bool, True), 
            "number_of_negatives_added_ratio": (float, 0.5), 
            "lower_bound_for_negatives": (int, 5), 
            "number_of_contradictory_answers_added_ratio": (float, 0.67), 
            "number_of_positives_added_ratio": (float, 0.0),
            "regenerate_embeddings": (float, True), 
            "synthetic_query_prompt": (str, "You are an expert question-answering system. You must create a question for the provided document. The question must be answerable within the context of the document.\n\n")
        },
        
        "classifier_model": {
            "training_dataset": (list, None), # Required parameter (No default)
            "validation_set": (list, None), # Required parameter (No default)
            "label_column": (list, None), # Required parameter (No default)
            "num_epochs": (int, None), # Required parameter (No default)
            "patience_value": (int, None), # Required parameter (No default)
            "learning_rate": (float, None), # Required parameter (No default)
            "training_dataset_path": (str, "None"),
            "validation_dataset_path": (str, "None"),
            "model_choice": (str, "microsoft/deberta-v3-large"), 
            "validation_set_scoring": (bool, True), 
            "assigned_batch_size": (int, REQUIRED_BUT_HAS_DEFAULT, 1),  # Default is 1
            "gradient_accumulation_multiplier": (int, REQUIRED_BUT_HAS_DEFAULT, 32),  # Default is 32
            "number_of_runs": (int, 1), 
            "num_warmup_steps": (int, 100),
            "training_row_limit": (int, -1),
            "validation_row_limit": (int, -1)
        },

        "ppi": {
            "evaluation_datasets": (list, None), # Required parameter (No default)
            "few_shot_examples_filepath": (str, None), # Required parameter (No default)
            "labels": (list, None), # Required parameter (No default)
            "gold_label_path": (str, None), # Required parameter (No default)
            "checkpoints": (list, []), # Required parameter (No default)
            "rag_type": (str, "question_answering"), 
            "annotated_datapoints_filepath": (bool, False), 
            "GPT_scoring":(bool, False), 
            "model_choice":(str, "microsoft/deberta-v3-large"),
            "llm_judge": (str, "None"), 
            "assigned_batch_size": (int, 1),
            "number_of_labels": (int, 2),
            "alpha": (int, 0.05),   
            "num_trials": (int, 1000),
            "vllm": (bool, False),
            "host_url": (str, "http://0.0.0.0:8000/v1"),
            "request_delay": (int, 0),
            "debug_mode": (bool, False)
        }
    }

    def __init__(self, synthetic_query_generator={}, ues_idp={}, classifier_model={}, ppi={}):
        self.synthetic_query_generator_config = self.prepare_config("synthetic_query_generator", synthetic_query_generator)
        self.classifier_model_config = self.prepare_config("classifier_model", classifier_model)
        self.ppi_config = self.prepare_config("ppi", ppi)
        self.ues_idp_config = self.prepare_config("ues_idp", ues_idp)

    def generate_synthetic_data(self): 
        if self.synthetic_query_generator_config == {}: 
            print("Skipping synthetic generator configuration due to missing parameters.")
        else: 
            synthetic_generator_config(**self.synthetic_query_generator_config) 

    def train_classifier(self): 
        if self.classifier_model_config == {}:
            print("Skipping binary classifier configuration due to missing parameters.")
        else:
            binary_classifer_config(**self.classifier_model_config) 

    def evaluate_RAG(self):
        if self.ppi_config == {}: 
            print("Skipping RAG evaluation configuration due to no parameters")
        else:
            rag_scoring_config(**self.ppi_config)
        
    def ues_idp(self):
        return ues_idp_config(**self.ues_idp_config)
        # if 'in_domain_prompts_dataset' in self.ues_idp_config:
        #     in_domain_prompts_dataset = pd.read_csv(self.ues_idp_config['in_domain_prompts_dataset'], sep='\t')
        # else:
        #     in_domain_prompts_dataset = None 

        # if 'unlabeled_evaluation_set' in self.ues_idp_config:
        #     unlabeled_evaluation_set = pd.read_csv(self.ues_idp_config['unlabeled_evaluation_set'], sep='\t')
        # else:
        #     unlabeled_evaluation_set = None
        # if self.ues_idp_config == {}: 
        #     print(f"Skipping UES and IDP since either unlabeled evaluation set or in domain prompts not provided")
        # else: 
        #     model_choice = self.ues_idp_config["model_choice"]
    
    def KILT_dataset(self, dataset_name): 
        KILT_dataset_process(dataset_name)

    def superGlue_dataset(self, dataset_name): 
        superGlue(dataset_name)

    def run(self): 
        if self.synthetic_query_generator_config == {}: 
            print("Skipping synthetic generator configuration due to no parameters.")
        else: 
            synthetic_generator_config(**self.synthetic_query_generator_config) 

        if self.classifier_model_config == {}:
            print("Skipping binary classifier configuration due to no parameters.")
        else:
            binary_classifer_config(**self.classifier_model_config) 
        
        if self.ppi_config == {}: 
            print("Skipping RAG evaluation configuration due to no parameters")
        else:
            rag_scoring_config(**self.ppi_config)
        
        # if 'in_domain_prompts_dataset' in self.ues_idp_config:
        #     in_domain_prompts_dataset = pd.read_csv(self.ues_idp_config['in_domain_prompts_dataset'], sep='\t')
        # else:
        #     in_domain_prompts_dataset = None 

        # if 'unlabeled_evaluation_set' in self.ues_idp_config:
        #     unlabeled_evaluation_set = pd.read_csv(self.ues_idp_config['unlabeled_evaluation_set'], sep='\t')
        # else:
        #     unlabeled_evaluation_set = None
        if self.ues_idp_config == {}: 
            print(f"Skipping UES and IDP since either unlabeled evaluation set or in domain prompts not provided")
            exit()
        else: 
            model_choice = self.ues_idp_config["model_choice"]
            ues_idp_config(**self.ues_idp_config)

    def prepare_config(self, component_name, user_config):
        if not user_config:
            return {}
        component = self.config_spec[component_name]
        prepared_config = {}
        for param, config in component.items():
            expected_type, default, default_value = config if len(config) == 3 else (*config, None)
            if param in user_config:
                value = user_config[param]
                if not isinstance(value, expected_type):
                    raise TypeError(f"Parameter '{param}' for {component_name} is expected to be of type {expected_type.__name__}, received {type(value).__name__} instead.")
                prepared_config[param] = value
            elif default is self.REQUIRED_BUT_HAS_DEFAULT:
                # If the parameter is required but not provided, use the specified default value.
                if param not in user_config:
                    print(f"\nWarning: '{param}' not provided for {component_name}, using default value {default_value}.")
                prepared_config[param] = default_value
            elif default is not None:
                prepared_config[param] = default
            else:
                raise ValueError(f"Missing required parameter '{param}' for {component_name}.")
        return prepared_config



from ragas import evaluate
from datasets import Dataset
import os
import pandas as pd
import scipy.stats as stats

from ragas.metrics import (
    answer_relevancy,
    faithfulness,
    context_recall,
    context_precision,
)

#os.environ["OPENAI_API_KEY"] = "your-openai-key"

evaluation_datasets = ['../datasets_v2/nq/ratio_0.7_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/nq/ratio_0.725_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/nq/ratio_0.75_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/nq/ratio_0.775_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/nq/ratio_0.8_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/nq/ratio_0.825_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/nq/ratio_0.85_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/nq/ratio_0.875_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/nq/ratio_0.9_reformatted_full_articles_False_validation_with_negatives.tsv']
#evaluation_datasets = ['../datasets_v2/hotpotqa/ratio_0.7_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/hotpotqa/ratio_0.725_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/hotpotqa/ratio_0.75_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/hotpotqa/ratio_0.775_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/hotpotqa/ratio_0.8_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/hotpotqa/ratio_0.825_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/hotpotqa/ratio_0.85_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/hotpotqa/ratio_0.875_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/hotpotqa/ratio_0.9_reformatted_full_articles_False_validation_with_negatives.tsv']
#evaluation_datasets = ['../datasets_v2/wow/ratio_0.7_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/wow/ratio_0.725_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/wow/ratio_0.75_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/wow/ratio_0.775_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/wow/ratio_0.8_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/wow/ratio_0.825_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/wow/ratio_0.85_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/wow/ratio_0.875_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/wow/ratio_0.9_reformatted_full_articles_False_validation_with_negatives.tsv']
#evaluation_datasets = ['../datasets_v2/fever/ratio_0.7_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/fever/ratio_0.725_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/fever/ratio_0.75_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/fever/ratio_0.775_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/fever/ratio_0.8_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/fever/ratio_0.825_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/fever/ratio_0.85_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/fever/ratio_0.875_reformatted_full_articles_False_validation_with_negatives.tsv', '../datasets_v2/fever/ratio_0.9_reformatted_full_articles_False_validation_with_negatives.tsv']
#evaluation_datasets = ['../datasets_v2/multirc/ratio_0.7_validation_with_negatives.tsv', '../datasets_v2/multirc/ratio_0.725_validation_with_negatives.tsv', '../datasets_v2/multirc/ratio_0.75_validation_with_negatives.tsv', '../datasets_v2/multirc/ratio_0.775_validation_with_negatives.tsv', '../datasets_v2/multirc/ratio_0.8_validation_with_negatives.tsv', '../datasets_v2/multirc/ratio_0.825_validation_with_negatives.tsv', '../datasets_v2/multirc/ratio_0.85_validation_with_negatives.tsv', '../datasets_v2/multirc/ratio_0.875_validation_with_negatives.tsv', '../datasets_v2/multirc/ratio_0.9_validation_with_negatives.tsv']
#evaluation_datasets = ['../datasets_v2/record/ratio_0.7_validation_with_negatives.tsv', '../datasets_v2/record/ratio_0.725_validation_with_negatives.tsv', '../datasets_v2/record/ratio_0.75_validation_with_negatives.tsv', '../datasets_v2/record/ratio_0.775_validation_with_negatives.tsv', '../datasets_v2/record/ratio_0.8_validation_with_negatives.tsv', '../datasets_v2/record/ratio_0.825_validation_with_negatives.tsv', '../datasets_v2/record/ratio_0.85_validation_with_negatives.tsv', '../datasets_v2/record/ratio_0.875_validation_with_negatives.tsv', '../datasets_v2/record/ratio_0.9_validation_with_negatives.tsv']

#evaluation_datasets = evaluation_datasets[:2]
correct_ranking = [i for i in range(0, len(evaluation_datasets))]

use_annotations_for_ranking = True

#labels = ["Context Relevance", "Answer Faithfulness", "Answer Relevance"]
labels = ["Context Relevance", "Answer Relevance"]

context_scores = []
answer_relevance_scores = []
answer_faithfulness_scores = []

for evaluation_dataset in evaluation_datasets:
    dataset = pd.read_csv(evaluation_dataset.replace("../", "../../ColBERT-FM/"), sep="\t")
    dataset = dataset[:2000]

    def string_to_list(text):
        return [text]

    # Apply the function to the 'text_column' to convert it to a column of lists
    dataset['contexts'] = dataset['Document'].apply(string_to_list)

    dataset = Dataset.from_pandas(dataset)
    dataset = dataset.rename_column("Query", "question")
    dataset = dataset.rename_column("Answer", "answer")
    #dataset = dataset.rename_column("Document", "contexts")

    if not use_annotations_for_ranking:

        results = evaluate(dataset, metrics=[
            context_precision,
            #faithfulness,
            answer_relevancy,
        ])

        print("Results for " + evaluation_dataset)
        print(results)

        context_scores.append(results['context_precision'])
        answer_relevance_scores.append(results['answer_relevancy'])
        #answer_faithfulness_scores.append(results['faithfulness'])

    else:
        
        sampled_y_labels = dataset.sample(n=300, random_state=42)
        context_relevance_prediction = sum(sampled_y_labels["Context_Relevance_Label"].tolist()) / len(sampled_y_labels)
        answer_relevance_prediction = sum(sampled_y_labels["Answer_Relevance_Label"].tolist()) / len(sampled_y_labels)
        context_scores.append(context_relevance_prediction)
        answer_relevance_scores.append(answer_relevance_prediction)

####################################

for label, scores in zip(labels, [context_scores, answer_relevance_scores]):

    indexed_list = list(enumerate(scores))
    sorted_list = sorted(indexed_list, key=lambda x: x[1])
    sorted_indices = [index for index, _ in sorted_list]
    tau, p_value = stats.kendalltau(correct_ranking, sorted_indices)

    print("--------------------------------------------------")
    print(label + " Scoring")
    print("Correct Ranking v. LLM-as-a-Judge Ranking")
    print(correct_ranking)
    print(sorted_indices)
    print("Kendall's Tau: " + str(tau))
    print("P-Value: " + str(p_value))
    print("Scores: " + str(scores))
    print("--------------------------------------------------")


# prepare your huggingface dataset in the format
# Dataset({
#     features: ['question', 'contexts', 'answer', 'ground_truths'],
#     num_rows: 25
# })

#dataset: Dataset

#results = evaluate(dataset)
# {'ragas_score': 0.860, 'context_precision': 0.817,
# 'faithfulness': 0.892, 'answer_relevancy': 0.874}




import requests
import time
import pandas as pd
import ast
import json
import copy
import openai
from tqdm import tqdm
import csv
from datasets import Dataset
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM, AutoConfig, AutoModelForCausalLM, BitsAndBytesConfig
import torch
import numpy as np
from sklearn.model_selection import train_test_split
import re
import argparse
import pdb
import sys
import warnings

from ares.LLM_as_a_Judge_Adaptation.LLM_Generation_Functions import generate_synthetic_query_openai_approach, generate_answer_from_context, generate_contradictory_answer_from_context
from ares.LLM_as_a_Judge_Adaptation.LLM_Generation_Functions import check_generated_answer, generate_contradictory_answer_examples, generate_synthetic_query_llm_approach, generate_answer_llm_approach

from ares.LLM_as_a_Judge_Adaptation.Filter_Synthetic_Queries import get_embedding, generate_index, filter_synthetic_queries, generate_additional_negatives, generate_additional_positives
#from Instruction_Finetune import instruction_finetune
#from Instruction_Finetune_V2 import instruction_finetune_v2
#from Instruction_Finetune_v3 import instruction_finetune_v3

pd.set_option('display.max_columns', None)  # Show all columns - TEST
pd.set_option('display.max_rows', None)  # Show all rows - TEST
pd.set_option('display.max_colwidth', None)  # Show full content of each column - TEST

#################################################

def clean_document(document: str):
    cleaned_document = re.sub(r'\n+', '\n', document.replace("\r"," ").replace("\t"," ")).strip()
    cleaned_document = cleaned_document.replace("=", " ").replace("-", " ")
    cleaned_document = re.sub(r'\s+', ' ', cleaned_document).strip()
    cleaned_document = (" ").join(cleaned_document.split(" ")) #[:512]
    return cleaned_document

    # Load model for synthetic query generation 

def validate_input_file(df, required_columns) -> bool: 
    missing_columns = [col for col in required_columns if col not in df.columns]
    if missing_columns:
        sys.exit(f"Error: The DataFrame is missing the following required column(s): {', '.join(missing_columns)}.")
############

def load_model(flan_approach, model_choice):
    if flan_approach:
        tokenizer = AutoTokenizer.from_pretrained(model_choice)
        model = AutoModelForSeq2SeqLM.from_pretrained(model_choice)

        torch.no_grad()
        model.eval()

        device = "cuda:0"
        device = torch.device(device)
        model.to(device)
    return model, tokenizer, device

#############

# Load documents for synthetic query and answer generation
def load_documents(document_filepath, clean_documents, documents_sampled):
    documents = []
    required_columns = ['Query', 'Document', 'Answer']

    if "docs_aws" in document_filepath:

        with open(document_filepath, "r") as json_file:
            json_data = json.load(json_file)
            documents = [x['text'] for x in json_data]

            # Clean document
            if clean_documents:
                documents = [clean_document(text) for text in documents]

        documents = pd.DataFrame(documents, columns=["document"])

    else:
        if not document_filepath.endswith('.tsv'):
            sys.exit(f"Error: The file {document_filepath} is not a TSV file.")
        try:
            documents = pd.read_csv(document_filepath, sep="\t")
            validate_input_file(documents, required_columns)
            documents.rename(columns={"Document": "document"}, inplace=True)
            documents['document'] = documents['document'].str.strip()
        except Exception as e:
            sys.exit(f"Error reading the file {document_filepath}: {e}")


    initial_count = len(documents)
    documents = documents[documents['document'].str.split().apply(len) >= 50] # Filter documents w/ less than 50 words.
    after_filter_count = len(documents)

    count = initial_count - after_filter_count

    if(after_filter_count == 0): 
        sys.exit(f"All documents were less than 50 words, please provide dataset with documents containing more than 50 words")

    if documents_sampled > initial_count:
        print(f"\nThe `documents_sampled` parameter ({documents_sampled}) exceeds the available number of documents ({initial_count}). Sampling will be adjusted to the maximum available documents ({initial_count}).\n")
        documents_sampled = initial_count

    if count > 0: 
        print(f"Filtered out {count} documents because they had less than 50 words. Sampling will be be adjusted to {after_filter_count} documents\n")
        documents_sampled = after_filter_count

    documents = documents.sample(n=documents_sampled, random_state=43)

    #documents = documents[:10]
    # print("documents") - CHANGED
    # print(len(documents)) - CHANGED
    # print(documents.head()) - CHANGED
    return documents

################

# Load few-shot prompt
def load_few_shot_prompt(few_shot_prompt_filename,for_fever_dataset,for_wow_dataset):
    few_shot_prompt = pd.read_csv(few_shot_prompt_filename, sep="\t")
    few_shot_prompt = few_shot_prompt[few_shot_prompt['Context_Relevance_Label'] == "[[Yes]]"]
    if "Query" not in few_shot_prompt:
        few_shot_prompt['Query'] = few_shot_prompt['Question']

    length_of_fewshot_prompt = len(few_shot_prompt)
    few_shot_examples = ""
    for row in range(len(few_shot_prompt)):
        few_shot_examples += "Example " + str(row + 1) + ":\n"
        few_shot_examples += "Document: " + clean_document(few_shot_prompt.iloc[row]['Document']) + "\n"
        if for_fever_dataset:
            few_shot_examples += "Statement: " + few_shot_prompt.iloc[row]['Query'] + "\n\n"
        elif for_wow_dataset:
            few_shot_examples += "Dialogue: " + few_shot_prompt.iloc[row]['Query'] + "\n\n"
        else:
            few_shot_examples += "Question: " + few_shot_prompt.iloc[row]['Query'] + "\n\n"

    # print("Fewshot Prompt") - CHANGED
    # print(few_shot_examples) - CHANGED 
    # print("Finished loading dataset + model") - CHANGED
    return few_shot_examples, length_of_fewshot_prompt

#################################################
def generate_contradictory_answers(few_shot_prompt_filename,generate_contradictory_answers_with_flan,for_fever_dataset,for_wow_dataset): 
    few_shot_prompt_for_contradictory_answers = pd.read_csv(few_shot_prompt_filename, sep="\t")
    few_shot_prompt_for_contradictory_answers = few_shot_prompt_for_contradictory_answers[few_shot_prompt_for_contradictory_answers['Contradictory_Answer'].str.len() > 4]

    if generate_contradictory_answers_with_flan:
        few_shot_examples_for_contradictory_answers = ""
        for row in range(len(few_shot_prompt_for_contradictory_answers)):
            few_shot_examples_for_contradictory_answers += "Example " + str(row + 1) +":\n"
            few_shot_examples_for_contradictory_answers += "Document: " + few_shot_prompt_for_contradictory_answers.iloc[row]['Document'] + "\n"
            if for_fever_dataset:
                few_shot_examples_for_contradictory_answers += "Statement: " + few_shot_prompt_for_contradictory_answers.iloc[row]['Query'] + "\n"
                few_shot_examples_for_contradictory_answers += "Incorrect Answer: " + few_shot_prompt_for_contradictory_answers.iloc[row]['Contradictory_Answer'] + "\n\n"
            elif for_wow_dataset:
                few_shot_examples_for_contradictory_answers += "Dialogue: " + few_shot_prompt_for_contradictory_answers.iloc[row]['Query'] + "\n"
                few_shot_examples_for_contradictory_answers += "Incorrect Response: " + few_shot_prompt_for_contradictory_answers.iloc[row]['Contradictory_Answer'] + "\n\n"
            else:
                few_shot_examples_for_contradictory_answers += "Question: " + few_shot_prompt_for_contradictory_answers.iloc[row]['Query'] + "\n"
                few_shot_examples_for_contradictory_answers += "Incorrect Answer: " + few_shot_prompt_for_contradictory_answers.iloc[row]['Contradictory_Answer'] + "\n\n"
    return few_shot_examples_for_contradictory_answers

def generate_few_shot_prompts(few_shot_prompt_filename,for_fever_dataset,for_wow_dataset):
    answer_gen_few_shot_prompt = pd.read_csv(few_shot_prompt_filename, sep="\t")
    answer_gen_few_shot_prompt = answer_gen_few_shot_prompt[answer_gen_few_shot_prompt['Answer_Relevance_Label'] == "[[Yes]]"]
    answer_gen_few_shot_prompt = answer_gen_few_shot_prompt[answer_gen_few_shot_prompt['Answer_Faithfulness_Label'] == "[[Yes]]"]
    #answer_gen_few_shot_prompt = answer_gen_few_shot_prompt[:4]
    length_of_fewshot_prompt_answer_gen = len(answer_gen_few_shot_prompt)
    if "Query" in answer_gen_few_shot_prompt.columns:
        answer_gen_few_shot_prompt['Question'] = answer_gen_few_shot_prompt['Query']

    answer_gen_few_shot_examples = ""
    for row in range(len(answer_gen_few_shot_prompt)):
        answer_gen_few_shot_examples += "Example " + str(row + 1) +":\n"
        answer_gen_few_shot_examples += "Document: " + answer_gen_few_shot_prompt.iloc[row]['Document'] + "\n"
        if for_fever_dataset:
            answer_gen_few_shot_examples += "Statement: " + answer_gen_few_shot_prompt.iloc[row]['Query'] + "\n"
            answer_gen_few_shot_examples += "Answer: " + answer_gen_few_shot_prompt.iloc[row]['Answer'] + "\n\n"
        elif for_wow_dataset:
            answer_gen_few_shot_examples += "Dialogue: " + answer_gen_few_shot_prompt.iloc[row]['Query'] + "\n"
            answer_gen_few_shot_examples += "Response: " + answer_gen_few_shot_prompt.iloc[row]['Answer'] + "\n\n"
        else:
            answer_gen_few_shot_examples += "Question: " + answer_gen_few_shot_prompt.iloc[row]['Query'] + "\n"
            answer_gen_few_shot_examples += "Answer: " + answer_gen_few_shot_prompt.iloc[row]['Answer'] + "\n\n"

    # print("answer_gen_few_shot_examples") - CHANGED
    # print(answer_gen_few_shot_examples) - CHANGED
    return answer_gen_few_shot_examples, length_of_fewshot_prompt_answer_gen

##############################

 # Save Synthetic Queries
def save_synthetic_queries(documents, regenerate_synth_questions, flan_approach, few_shot_examples, 
    length_of_fewshot_prompt, device, tokenizer, model, percentiles, for_fever_dataset, for_wow_dataset, 
    synthetic_query_prompt, synthetic_queries_filename, question_temperatures): 

    # print("documents") - CHANGED 
    # print(len(documents)) - CHANGED
    # print(documents.head()) - CHANGED

    if regenerate_synth_questions:
        print("Beginning synthetic query generation!")
        tqdm.pandas(desc="Generating synthetic queries...", total=documents.shape[0])
        if flan_approach:
            # Testing
            # subset_documents = documents.head(10)  #
            # subset_documents["synthetic_query"] = subset_documents.apply(lambda x: generate_synthetic_query_llm_approach(x["document"], few_shot_examples, length_of_fewshot_prompt, device, tokenizer, model, percentiles, for_fever_dataset=for_fever_dataset, for_wow_dataset=for_wow_dataset), axis=1)
            # print(subset_documents[["document", "synthetic_query"]])
            # breakpoint()
            # synthetic_query = generate_synthetic_query_llm_approach(test_doc["document"], few_shot_examples, length_of_fewshot_prompt, device, tokenizer, model, percentiles, for_fever_dataset=for_fever_dataset, for_wow_dataset=for_wow_dataset) #
            # print(synthetic_query)
            documents["synthetic_query"] = documents.progress_apply(lambda x: generate_synthetic_query_llm_approach(x["document"], few_shot_examples, length_of_fewshot_prompt, device, tokenizer, model, percentiles, for_fever_dataset=for_fever_dataset, for_wow_dataset=for_wow_dataset), axis=1)
        else:
            documents["synthetic_query"] = documents.progress_apply(lambda x: generate_synthetic_query_openai_approach(x["document"], synthetic_query_prompt, few_shot_examples, question_temperatures, length_of_fewshot_prompt), axis=1)
        documents = documents.explode("synthetic_query", ignore_index=True)
        documents = documents.drop_duplicates(subset=['synthetic_query'])
        documents.to_csv(synthetic_queries_filename, index=False, sep="\t")
        print("Saved synthetic queries to: " + synthetic_queries_filename)

################################################

# Generate synthetic queries and their answers for training set
def Generate_Synthetic_Answers(synthetic_queries_filename,
    regenerate_answers, flan_approach, answer_gen_few_shot_examples, length_of_fewshot_prompt_answer_gen, 
    device, tokenizer, model, for_fever_dataset, for_wow_dataset, generate_contradictory_answers_with_flan,
    few_shot_examples_for_contradictory_answers, number_of_negatives_added_ratio,
    lower_bound_for_negatives, number_of_contradictory_answers_added_ratio, number_of_positives_added_ratio, regenerate_embeddings): 
    synth_queries = pd.read_csv(synthetic_queries_filename, sep="\t")
    # print("Filtered synth queries")
    # print(len(synth_queries))
    synth_queries = synth_queries[synth_queries["synthetic_query"].str.len() > 10]
    # print(len(synth_queries))

    if regenerate_answers:
        print("Beginning answer generation!")
        
        tqdm.pandas(desc="Generating answers...", total=synth_queries.shape[0])
        if flan_approach:
            synth_queries["generated_answer"] = synth_queries.progress_apply(lambda x: generate_answer_llm_approach(x["document"], x["synthetic_query"], answer_gen_few_shot_examples, length_of_fewshot_prompt_answer_gen, device, tokenizer, model, for_fever_dataset=for_fever_dataset, for_wow_dataset=for_wow_dataset), axis=1)
        else:
            synth_queries["generated_answer"] = synth_queries.progress_apply(lambda x: generate_answer_from_context(x["document"], x["synthetic_query"]), axis=1)
        synth_queries.to_csv(synthetic_queries_filename, index=False, sep="\t")
        print("Saved answers to: " + synthetic_queries_filename)

        #################################################

        #  for idx, row in synth_queries.iterrows():
        #     print(f"Document: {row['document']}\nSynthetic Query: {row['synthetic_query']}\nGenerated Answer: {row['generated_answer']}\n---\n")
        # breakpoint()

        synth_queries["Answer_Faithfulness_Label"] = [check_generated_answer(synth_queries.iloc[i]['generated_answer']) for i in range(len(synth_queries))]
        synth_queries["Answer_Relevance_Label"] = [check_generated_answer(synth_queries.iloc[i]['generated_answer']) for i in range(len(synth_queries))]
        print("Generating contradictory answers!")
        if generate_contradictory_answers_with_flan:
            synth_queries = generate_contradictory_answer_examples(synth_queries, int(len(synth_queries) * number_of_contradictory_answers_added_ratio), few_shot_examples_for_contradictory_answers=few_shot_examples_for_contradictory_answers, device=device, tokenizer=tokenizer, model=model, for_fever_dataset=for_fever_dataset, for_wow_dataset=for_wow_dataset)
        else:
            synth_queries = generate_contradictory_answer_examples(synth_queries, int(len(synth_queries) * number_of_contradictory_answers_added_ratio))
        # print(synth_queries.columns)  # List all column names to understand the structure
        # print(synth_queries.head()) 
        synth_queries.to_csv(synthetic_queries_filename, index=False, sep="\t")
        print("Saved answers to: " + synthetic_queries_filename)

    #################################################

    synth_queries = pd.read_csv(synthetic_queries_filename, sep="\t")
    synth_queries = synth_queries[synth_queries["synthetic_query"].str.len() > 10]

    # print("synth_queries")
    # print(len(synth_queries))
    # print(synth_queries.head())
    

    if regenerate_embeddings:
        print("Generating index and negatives!")
        documentation_index = generate_index(synth_queries)
        synth_queries = filter_synthetic_queries(synth_queries, documentation_index)
        synth_queries = generate_additional_negatives(synth_queries, documentation_index, number_of_negatives_added_ratio, lower_bound_for_negatives)
        synth_queries = generate_additional_positives(synth_queries, documentation_index, number_of_positives_added_ratio)
        synth_queries.to_csv(synthetic_queries_filename, index=False, sep="\t")

    #################################################

    synth_queries = pd.read_csv(synthetic_queries_filename, sep="\t")

    #Shuffle queries
    synth_queries = synth_queries.sample(n=len(synth_queries), random_state=42)

    # print("Label Sets for each Metric")
    # print(set(synth_queries['Context_Relevance_Label'].tolist()))
    # print(set(synth_queries['Answer_Faithfulness_Label'].tolist()))
    # print(set(synth_queries['Answer_Relevance_Label'].tolist()))

    # print("synth_queries filtered")
    # print(len(synth_queries))
    # print(synth_queries.head())

    # print("Positive and Negative Counts")
    # print("Context Relevance")
    # print(len(synth_queries[synth_queries['Context_Relevance_Label'] == "Yes"]))
    # print(len(synth_queries[synth_queries['Context_Relevance_Label'] == "No"]))
    # print("Answer Faithfulness")
    # print(len(synth_queries[synth_queries['Answer_Faithfulness_Label'] == "Yes"]))
    # print(len(synth_queries[synth_queries['Answer_Faithfulness_Label'] == "No"]))
    # print("Answer Relevance")
    # print(len(synth_queries[synth_queries['Answer_Relevance_Label'] == "Yes"]))
    # print(len(synth_queries[synth_queries['Answer_Relevance_Label'] == "No"]))

    synth_queries.to_csv(synthetic_queries_filename, index=False, sep="\t")
    print("Completed synthetic generation!")
    print("Saved synthetic queries file to: " + synthetic_queries_filename)

# def print_synthetic_queries(filename): 
#     queries = pd.read_csv(filename,sep='\t')
    
#     for index, row in queries.head(6).iterrows(): 
#         print(f"Document {index}: {row['document']}")
#         print(f"Synthetic Query: {row['synthetic_query']}")
#         print(f"Answer: {row['generated_answer']}")
#         print("-" * 50)

# def synthetic_generator_config(document_filepath: str, few_shot_prompt_filename: str,
#                                synthetic_queries_filename: str, documents_sampled: int,
#                                flan_approach: bool = True, clean_documents: bool = False,
#                                regenerate_synth_questions: bool = True, 
#                                percentiles: list = [0.05, 0.25, 0.5, 0.95], 
#                                question_temperatures: list = [2.0, 1.5, 1.0, 0.5, 0.0],
#                                regenerate_answers: bool = True,
#                                generate_contradictory_answers_with_flan: bool = True, 
#                                number_of_negatives_added_ratio: float = 0.5, # Check whether can also be an int
#                                lower_bound_for_negatives: int = 5, # Need to be an int value
#                                number_of_contradictory_answers_added_ratio: float = 0.67, # Check whether can also be an int
#                                number_of_positives_added_ratio: float = 0.0, # Check whether can also be an int
#                                regenerate_embeddings: float = True, 
#                                synthetic_query_prompt: str = "You are an expert question-answering system. You must create a question for the provided document. The question must be answerable within the context of the document.\n\n"
#                                ): 
#     for_fever_dataset = False
#     if "fever" in document_filepath.lower():
#         for_fever_dataset = True
#     for_wow_dataset = False
#     if "wow" in document_filepath.lower():
#         for_wow_dataset = True

#     model, model_choice, tokenizer, device = load_model(flan_approach)

#     documents = load_documents(document_filepath, clean_documents, documents_sampled)

#     few_shot_examples, length_of_fewshot_prompt = load_few_shot_prompt(few_shot_prompt_filename,for_fever_dataset,for_wow_dataset)

#     few_shot_examples_for_contradictory_answers = generate_contradictory_answers(few_shot_prompt_filename,generate_contradictory_answers_with_flan,for_fever_dataset,for_wow_dataset)

#     answer_gen_few_shot_examples, length_of_fewshot_prompt_answer_gen = generate_few_shot_prompts(few_shot_prompt_filename,for_fever_dataset,for_wow_dataset)

#     save_synthetic_queries(documents, regenerate_synth_questions, flan_approach, few_shot_examples, 
#     length_of_fewshot_prompt, device, tokenizer, model, percentiles, for_fever_dataset, for_wow_dataset, 
#     synthetic_query_prompt, synthetic_queries_filename, question_temperatures)

#     Generate_Synthetic_Queries_and_Answers(synthetic_queries_filename,
#     regenerate_answers, flan_approach, answer_gen_few_shot_examples, length_of_fewshot_prompt_answer_gen, 
#     device, tokenizer, model, for_fever_dataset, for_wow_dataset, generate_contradictory_answers_with_flan,
#     few_shot_examples_for_contradictory_answers, number_of_negatives_added_ratio, lower_bound_for_negatives, number_of_contradictory_answers_added_ratio, number_of_positives_added_ratio, regenerate_embeddings)

#     print_synthetic_queries(synthetic_queries_filename)

#################################################

if __name__ == '__main__':

    parser = argparse.ArgumentParser()

    parser.add_argument("--document_filepath", type=str, required=True)
    parser.add_argument("--few_shot_prompt_filename", type=str, required=True)
    parser.add_argument("--synthetic_queries_filename", type=str, required=True)
    parser.add_argument("--flan_approach", type=bool, default="True", required=False)
    parser.add_argument("--documents_sampled", type=int, required=True)

    args = parser.parse_args()

    ### Instructions

    document_filepath = args.document_filepath
    few_shot_prompt_filename = args.few_shot_prompt_filename
    synthetic_queries_filename = args.synthetic_queries_filename
    flan_approach = args.flan_approach
    if flan_approach == "True":
        flan_approach = True 
    else:
        flan_approach = False
    documents_sampled = args.documents_sampled

    ########################################################

    for_fever_dataset = False
    if "fever" in document_filepath.lower():
        for_fever_dataset = True
    for_wow_dataset = False
    if "wow" in document_filepath.lower():
        for_wow_dataset = True

    regenerate_synth_questions = True
    regenerate_answers = True
    regenerate_embeddings = True

    lower_bound_for_negatives = 20
    number_of_negatives_added_ratio = 0.5
    number_of_positives_added_ratio = 0.0 #0.2
    number_of_contradictory_answers_added_ratio = 0.67
    clean_documents = False

    question_temperatures = [2.0, 1.5, 1.0, 0.5, 0.0] #[2.0, 0.9, 0.7, 0.5]
    percentiles = [0.05, 0.25, 0.5, 0.95] #[0.01, 0.05, 0.25, 0.5, 0.95]
    clean_few_shot_prompt_docs = False

    flan_approach = True
    generate_contradictory_answers_with_flan = True

    #################################################

    synthetic_query_prompt = "You are an expert question-answering system. You must create a question for the provided document. The question must be answerable within the context of the document.\n\n"

    print("------------------------------------------------------------")
    print("Document File: " + document_filepath)
    print("Synthetic File Path: " + synthetic_queries_filename)
    print("number_of_negatives_added_ratio: " + str(number_of_negatives_added_ratio))
    print("number_of_positives_added_ratio: " + str(number_of_positives_added_ratio))
    print("number_of_contradictory_answers_added_ratio: " + str(number_of_contradictory_answers_added_ratio))
    print("clean_documents: " + str(clean_documents))
    print("question_temperatures: " + str(question_temperatures))
    print("percentiles: " + str(percentiles))
    print("lower_bound_for_negatives: " + str(lower_bound_for_negatives))
    print("for_fever_dataset: " + str(for_fever_dataset))
    print("for_wow_dataset: " + str(for_wow_dataset))
    print("------------------------------------------------------------")

    # model, model_choice, tokenizer, device = load_model(flan_approach)

    # documents = load_documents(document_filepath, clean_documents)

    # few_shot_examples, length_of_fewshot_prompt = load_few_shot_prompt(few_shot_prompt_filename,for_fever_dataset)

    # few_shot_examples_for_contradictory_answers = generate_contradictory_answers(few_shot_prompt_filename)

    # answer_gen_few_shot_examples, length_of_fewshot_prompt_answer_gen = generate_few_shot_prompts(few_shot_prompt_filename)

    # save_synthetic_queries(documents, regenerate_synth_questions, flan_approach, few_shot_examples, 
    # length_of_fewshot_prompt, device, tokenizer, model, percentiles, for_fever_dataset, for_wow_dataset, 
    # synthetic_query_prompt, synthetic_queries_filename, question_temperatures)

    # Generate_Synthetic_Queries_and_Answers(synthetic_queries_filename,
    # regenerate_answers, flan_approach, answer_gen_few_shot_examples, length_of_fewshot_prompt_answer_gen, 
    # device, tokenizer, model, for_fever_dataset, for_wow_dataset, generate_contradictory_answers_with_flan,
    # number_of_contradictory_answers_added_ratio, few_shot_examples_for_contradictory_answers, number_of_negatives_added_ratio
    # lower_bound_for_negatives, number_of_contradictory_answers_added_ratio, number_of_positives_added_ratio, regenerate_embeddings): 

    ######################################################################



import requests
import time
import pandas as pd
import ast
import json
import copy
import openai
from openai import OpenAI
from tqdm import tqdm
import csv
from datasets import Dataset
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
import torch
import numpy as np
import random
import pdb
import warnings
from pandas.errors import SettingWithCopyWarning


#################################################
client = OpenAI()

def get_embedding(text, model="text-embedding-ada-002"):
    text = text.replace("\n", " ")
    if len(text) > 50:
        text = (" ").join(text.split(" ")[:50])
    for _ in range(5):
        try:
            return client.embeddings.create(input = [text], model=model).data[0].embedding
        except:
            print("Error generating embedding! Attempting again...")
            time.sleep(30)

def generate_index(dataframe):
    warnings.simplefilter("ignore", SettingWithCopyWarning)
    dataframe = dataframe.drop_duplicates(subset="document")
    tqdm.pandas(desc="Generating embeddings...", total=dataframe.shape[0])
    dataframe['embeddings'] = dataframe["document"].progress_apply(lambda x: get_embedding(x, model='text-embedding-ada-002')) # model='text-embedding-ada-002'
    dataframe =  dataframe[dataframe['embeddings'].apply(lambda x: len(x)) == 1536]
    
    dataframe = Dataset.from_pandas(dataframe)
    dataframe.add_faiss_index(column="embeddings")
    return dataframe

def filter_synthetic_queries(queries_dataset, document_index):
    
    total_filtered_questions = []
    total_labels = []
    
    queries_dataset = Dataset.from_pandas(queries_dataset)
    for i in tqdm(range(len(queries_dataset))):
        question = queries_dataset[i]["synthetic_query"]
        question_embedding = np.array(get_embedding(question)).astype(np.float32)
        scores, samples = document_index.get_nearest_examples("embeddings", question_embedding, k=20)
        if samples["document"][0] == queries_dataset[i]["document"]:
            total_labels.append("Yes")
        else:
            # Create negatives by filtering out top-k results 
            found_gold_in_top_k = False
            for j in range(20):
                found_gold_in_top_k = found_gold_in_top_k or (samples["document"][j] == queries_dataset[i]["document"])
            if not found_gold_in_top_k:
                total_labels.append("No")
            else:
                total_labels.append("N/A")

    #################################################################

    queries_dataset = queries_dataset.to_pandas()
    queries_dataset['Context_Relevance_Label'] = total_labels
    
    # print("Before filter")
    # print(len(queries_dataset))
    # print("After filter")
    # print(len(queries_dataset[queries_dataset['Context_Relevance_Label'] == "Yes"]))
    # print(len(queries_dataset[queries_dataset['Context_Relevance_Label'] == "No"]))

    return queries_dataset

############################################################

def generate_additional_negatives(queries_dataset, document_index, number_of_negatives_added_ratio: float, lower_bound_for_negatives: int):
    
    negative_documents = []
    negative_labels = []
    queries_dataset_copy = queries_dataset.copy().drop_duplicates(["synthetic_query", "document"])
    queries_dataset_copy = queries_dataset_copy.sample(n=int(len(queries_dataset_copy) * number_of_negatives_added_ratio), random_state=42)
    negative_sample_retrieved = []

    for i in tqdm(range(len(queries_dataset_copy))):
        question = queries_dataset_copy.iloc[i]["synthetic_query"]
        question_embedding = np.array(get_embedding(question)).astype(np.float32)
        scores, samples = document_index.get_nearest_examples(
            "embeddings", question_embedding, k=100
        )
        # if len(samples["document"]) <= 98:
        # print(f"Number of samples: {len(samples['document'])}")
        #    raise ValueError('Less than 100 documents in dataset! Please add more documents for retrieval.')
        random_negative_sample = random.randint(lower_bound_for_negatives, len(samples["document"]) - 1)
        negative_sample_retrieved.append(random_negative_sample)
        try:
            negative_documents.append(samples["document"][random_negative_sample])
        except:
            negative_documents.append(samples["document"][0])
        negative_labels.append("No")

    # Swap documents with negative documents + negative labels
    queries_dataset_copy["document"] = negative_documents
    queries_dataset_copy['Context_Relevance_Label'] = negative_labels
    queries_dataset_copy['negative_sample_retrieved'] = negative_sample_retrieved

    # print("Negatives Added")
    # print(len(queries_dataset_copy))
    
    # Shuffle dataframe
    queries_dataset_copy = queries_dataset_copy.sample(n=len(queries_dataset_copy), random_state=42)

    queries_dataset = pd.concat([queries_dataset, queries_dataset_copy], axis=0, ignore_index=True)
    return queries_dataset

############################################################

def generate_additional_positives(queries_dataset, document_index, number_of_positives_added_ratio: float):
    
    positive_queries = []
    positive_documents = []
    positive_labels = []
    
    queries_dataset_copy = queries_dataset.copy()
    queries_dataset_copy = queries_dataset_copy.drop_duplicates(["synthetic_query", "document"])
    queries_dataset_copy = queries_dataset_copy[queries_dataset_copy['Context_Relevance_Label'] == "Yes"]
    queries_dataset_copy = queries_dataset_copy.sample(n=int(len(queries_dataset_copy) * number_of_positives_added_ratio), random_state=42)

    for i in tqdm(range(len(queries_dataset_copy))):
        question = queries_dataset_copy.iloc[i]["synthetic_query"]
        question_embedding = np.array(get_embedding(question)).astype(np.float32)
        scores, samples = document_index.get_nearest_examples(
            "embeddings", question_embedding, k=100
        )

    # Swap documents with positive docs + labels
    if len(positive_queries) < len(queries_dataset_copy):
        queries_dataset_copy = queries_dataset_copy[:len(positive_queries)]
    queries_dataset_copy["synthetic_query"] = positive_queries[:len(queries_dataset_copy)]
    queries_dataset_copy["document"] = positive_documents[:len(queries_dataset_copy)]
    queries_dataset_copy['Context_Relevance_Label'] = positive_labels[:len(queries_dataset_copy)]
    queries_dataset_copy['generated_answer'] = ["" for j in range(len(queries_dataset_copy))]
    queries_dataset_copy['additional_positive_added'] = [True for j in range(len(queries_dataset_copy))]

    # print("Positives Added")
    # print(len(queries_dataset_copy))

    # Shuffle dataframe
    queries_dataset_copy = queries_dataset_copy.sample(n=len(queries_dataset_copy), random_state=42)

    queries_dataset = pd.concat([queries_dataset, queries_dataset_copy], axis=0, ignore_index=True)
    return queries_dataset






import requests
import time
import pandas as pd
import ast
import json
import copy
import openai
from openai import OpenAI
from tqdm import tqdm
import csv
from datasets import Dataset
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM, AutoConfig, AutoModelForCausalLM
from vllm import LLM
import torch
import numpy as np
from sklearn.model_selection import train_test_split
import requests 
import os
import re
import time
import anthropic
import sys
import re
import warnings

####################################################################

def calculate_accuracy(predictions, ground_truth):
    if len(predictions) != len(ground_truth):
        print(len(predictions))
        print(len(ground_truth))
        raise ValueError("Input lists must have the same length")

    correct_count = sum(1 for pred, truth in zip(predictions, ground_truth) if pred == truth)
    total_count = len(predictions)

    accuracy = round(correct_count * 100 / total_count, 2)
    return accuracy

####################################################################

def few_shot_context_relevance_scoring(system_prompt: str, query: str, document: str, gpt_model: str, query_id: str, debug_mode: bool, request_delay: int, few_shot_examples=None):
    for _ in range(5):
        #try:

            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += "Question: " + few_shot_examples.iloc[row][query_id] + "\n"
                    user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
                    current_label = few_shot_examples.iloc[row]['Context_Relevance_Label']
                    if current_label == 1 or current_label == 1.0:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label == 0 or current_label == 0.0:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"
            
            user_prompt += "Question: " + query + "\n"
            user_prompt += "Document: " + document + "\n"
            user_prompt += "Label: "

            time.sleep(request_delay)

            if debug_mode is True: 
                    print("------------------------------------------")
                    print("Context Relevance")
                    print(user_prompt)
                    print("\n")
                    print("Question:", query)
                    print("\n")
                    print("Document:", document)
                    print("\n")
                    print("Label:")

            

            messages = [
                            {
                                "role": "system",
                                "content": system_prompt
                            },
                            {
                                "role": "user",
                                "content": user_prompt
                            }
                        ]
            response = openai.chat.completions.create(
                                model=gpt_model,
                                messages=messages
                            )    
            final_response = response.choices[0].message.content

            if debug_mode is True: 
                print(final_response)

            yes = r"\[\[Yes]]"
            no = r"\[\[No]]"
    
            if re.search(yes, final_response): 
                if debug_mode is True:
                    print("Returned label 1")
                return 1
            elif re.search(no, final_response):
                if debug_mode is True:
                    print("Returned label 0")
                return 0
            else:
                print("Didn't extract Yes or No!")
                # print(f"Unexpected response received: {final_response}")
                return -1
    
####################################################################

def few_shot_answer_faithfulness_scoring(system_prompt, query: str, document: str, answer: str, gpt_model: str, query_id: str, debug_mode: bool, request_delay: int, few_shot_examples=None):

    for _ in range(5):
        #try:

            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += "Question: " + few_shot_examples.iloc[row][query_id] + "\n"
                    user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
                    user_prompt += "Answer: " + few_shot_examples.iloc[row]['Answer'] + "\n"
                    current_label = few_shot_examples.iloc[row]['Answer_Faithfulness_Label']
                    if current_label == 1 or current_label == 1.0:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label == 0 or current_label == 0.0:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"
            
            user_prompt += "Question: " + query + "\n"
            user_prompt += "Document: " + document + "\n"
            user_prompt += "Answer: " + str(answer) + "\n"
            user_prompt += "Label: "

            time.sleep(request_delay)
            
            if debug_mode is True: 
                print("------------------------------------------")
                print("Answer Faithfulness")
                print(user_prompt)
                print("\n")
                print("Question:", query)
                print("\n")
                print("Document:", document)
                print("\n")
                print("Answer:", answer)
                print("\n")
                print("Label:")

            messages = [
                            {
                                "role": "system",
                                "content": system_prompt
                            },
                            {
                                "role": "user",
                                "content": user_prompt
                            }
                        ]
            response = openai.chat.completions.create(
                                model=gpt_model,
                                messages=messages
                            )    
            final_response = response.choices[0].message.content

            if debug_mode is True: 
                print(final_response)

            yes = r"\[\[Yes]]"
            no = r"\[\[No]]"
    
            if re.search(yes, final_response): 
                if debug_mode is True:
                    print("Returned label 1")
                return 1
            elif re.search(no, final_response):
                if debug_mode is True:
                    print("Returned label 0")
                return 0
            else:
                print("Didn't extract Yes or No!")
                # print(f"Unexpected response received: {final_response}")
                return -1
        #except:
            print("Error with querying OpenAI! Trying again...")
            time.sleep(60)
    
####################################################################

def few_shot_answer_relevance_scoring(system_prompt: str, query: str, document: str, answer: str, gpt_model: str, query_id: str, debug_mode: bool, request_delay: int, few_shot_examples=None):

    for _ in range(5):
        #try:

            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += "Question: " + few_shot_examples.iloc[row][query_id] + "\n"
                    user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
                    user_prompt += "Answer: " + few_shot_examples.iloc[row]['Answer'] + "\n"
                    current_label = few_shot_examples.iloc[row]['Answer_Relevance_Label']
                    if current_label == 1 or current_label == 1.0:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label == 0 or current_label == 0.0:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"

            user_prompt += "Question: " + query + "\n"
            user_prompt += "Document: " + document + "\n"
            user_prompt += "Answer: " + str(answer) + "\n"
            user_prompt += "Label: "

            time.sleep(request_delay)

            if debug_mode is True: 
                print("------------------------------------------")
                print("Answer Relevance")
                print(user_prompt)
                print("\n")
                print("Question:", query)
                print("\n")
                print("Document:", document)
                print("\n")
                print("Answer:", answer)
                print("\n")
                print("Label:")
            
            messages = [
                            {
                                "role": "system",
                                "content": system_prompt
                            },
                            {
                                "role": "user",
                                "content": user_prompt
                            }
                        ]
            response = openai.chat.completions.create(
                                model=gpt_model,
                                messages=messages
                            )    
            final_response = response.choices[0].message.content

            if debug_mode is True: 
                print(final_response)


            yes = r"\[\[Yes]]"
            no = r"\[\[No]]"
    
            if re.search(yes, final_response): 
                if debug_mode is True:
                    print("Returned label 1")
                return 1
            elif re.search(no, final_response):
                if debug_mode is True:
                    print("Returned label 0")
                return 0
            else:
                print("Didn't extract Yes or No!")
                # print(f"Unexpected response received: {final_response}")
                return -1
        #except:
            print("Error with querying OpenAI! Trying again...")
            time.sleep(60)


##############################TOGETHERAI CUSTOM MODELS (NOT OPENAI)######################################
            
def few_shot_context_relevance_scoring_togetherai(system_prompt: str, query: str, document: str, model_choice, query_id: str, debug_mode: bool, request_delay: int, few_shot_examples=None):
    TOGETHER_API_KEY = os.getenv("TOGETHER_API_KEY")

    if not TOGETHER_API_KEY:
        sys.exit("TOGETHER_API_KEY environment variable not set. Please set the variable.")

    user_prompt = ""
    if few_shot_examples is not None:
        for row in range(len(few_shot_examples)):
                    user_prompt += "Question: " + few_shot_examples.iloc[row][query_id] + "\n"
                    user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
                    current_label = few_shot_examples.iloc[row]['Context_Relevance_Label']
                    if current_label == 1 or current_label == 1.0:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label == 0 or current_label == 0.0:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"

    user_prompt += "Question: " + query + "\n"
    user_prompt += "Document: " + document + "\n"
    user_prompt += "Label: "

    time.sleep(request_delay)

    if debug_mode is True: 
        print("------------------------------------------")
        print("Context Relevance")
        print(user_prompt)
        print("\n")
        print("Question:", query)
        print("\n")
        print("Document:", document)
        print("\n")
        print("Label:")


    client = OpenAI(
    api_key=TOGETHER_API_KEY,
    base_url='https://api.together.xyz/v1',
    )

    chat_completion = client.chat.completions.create(
    messages=[
        {
        "role": "system",
        "content": system_prompt,
        },
        {
        "role": "user",
        "content": user_prompt,
        }
    ],
    model=model_choice
    )
    final_response = chat_completion.choices[0].message.content

    if debug_mode is True: 
        print(final_response)

    yes = r"\[\[Yes]]"
    no = r"\[\[No]]"
    
    if re.search(yes, final_response): 
        if debug_mode is True:
            print("Returned label 1")
        return 1
    elif re.search(no, final_response):
        if debug_mode is True:
            print("Returned label 0")
        return 0
    else:
        print("Didn't extract Yes or No!")
        # print(f"Unexpected response received: {final_response}")
        return -1
    # except requests.exceptions.RequestException as e:
    #     print(f"HTTP Request failed: {e}")
    #     return -1

def few_shot_answer_faithfulness_scoring_togetherai(system_prompt: str, query: str, document: str, answer: str, model_choice, query_id: str, debug_mode: bool, request_delay: int, few_shot_examples=None):
    TOGETHER_API_KEY = os.getenv("TOGETHER_API_KEY")

    if not TOGETHER_API_KEY:
        sys.exit("TOGETHER_API_KEY environment variable not set. Please set the variable.")

    user_prompt = ""
    if few_shot_examples is not None:
        for row in range(len(few_shot_examples)):
                user_prompt += "Question: " + few_shot_examples.iloc[row][query_id] + "\n"
                user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
                user_prompt += "Answer: " + few_shot_examples.iloc[row]['Answer'] + "\n"
                current_label = few_shot_examples.iloc[row]['Answer_Faithfulness_Label']
                if current_label == 1 or current_label == 1.0:
                    warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                    current_label = "[[Yes]]"
                elif current_label == 0 or current_label == 0.0:
                    warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                    current_label = "[[No]]"
                user_prompt += "Label: " + str(current_label) + "\n\n"

    user_prompt += "Question: " + query + "\n"
    user_prompt += "Document: " + document + "\n"
    user_prompt += "Answer: " + str(answer) + "\n"
    user_prompt += "Label: "

    time.sleep(request_delay)

    if debug_mode is True: 
        print("------------------------------------------")
        print("Answer Faithfulness")
        print(user_prompt)
        print("\n")
        print("Question:", query)
        print("\n")
        print("Document:", document)
        print("\n")
        print("Answer:", answer)
        print("\n")
        print("Label:")

    client = OpenAI(
    api_key=TOGETHER_API_KEY,
    base_url='https://api.together.xyz/v1',
    )

    chat_completion = client.chat.completions.create(
    messages=[
        {
        "role": "system",
        "content": system_prompt,
        },
        {
        "role": "user",
        "content": user_prompt,
        }
    ],
    model=model_choice
    )
    final_response = chat_completion.choices[0].message.content

    if debug_mode is True: 
        print(final_response)

    yes = r"\[\[Yes]]"
    no = r"\[\[No]]"
    
    if re.search(yes, final_response): 
        if debug_mode is True:
            print("Returned label 1")
        return 1
    elif re.search(no, final_response):
        if debug_mode is True:
            print("Returned label 0")
        return 0
    else:
        print("Didn't extract Yes or No!")
        # print(f"Unexpected response received: {final_response}")
        return -1
    # except requests.exceptions.RequestException as e:
    #     print(f"HTTP Request failed: {e}")
    #     return -1

def few_shot_answer_relevance_scoring_togetherai(system_prompt: str, query: str, document: str, answer: str, model_choice, query_id: str, debug_mode: bool, request_delay: int, few_shot_examples=None):
    TOGETHER_API_KEY = os.getenv("TOGETHER_API_KEY")

    if not TOGETHER_API_KEY:
        sys.exit("TOGETHER_API_KEY environment variable not set. Please set the variable.")

    user_prompt = ""
    if few_shot_examples is not None:
        for row in range(len(few_shot_examples)):
            user_prompt += "Question: " + few_shot_examples.iloc[row][query_id] + "\n"
            user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
            user_prompt += "Answer: " + few_shot_examples.iloc[row]['Answer'] + "\n"
            current_label = few_shot_examples.iloc[row]['Answer_Relevance_Label']
            if current_label == 1 or current_label == 1.0:
                warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                current_label = "[[Yes]]"
            elif current_label == 0 or current_label == 0.0:
                warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                current_label = "[[No]]"
            user_prompt += "Label: " + str(current_label) + "\n\n"

    user_prompt += "Question: " + query + "\n"
    user_prompt += "Document: " + document + "\n"
    user_prompt += "Answer: " + str(answer) + "\n"
    user_prompt += "Label: "

    time.sleep(request_delay)

    if debug_mode is True: 
        print("------------------------------------------")
        print("Answer Relevance")
        print(user_prompt)
        print("\n")
        print("Question:", query)
        print("\n")
        print("Document:", document)
        print("\n")
        print("Answer:", answer)
        print("\n")
        print("Label:")


    client = OpenAI(
    api_key=TOGETHER_API_KEY,
    base_url='https://api.together.xyz/v1',
    )

    chat_completion = client.chat.completions.create(
    messages=[
        {
        "role": "system",
        "content": system_prompt,
        },
        {
        "role": "user",
        "content": user_prompt,
        }
    ],
    model=model_choice
    )
    final_response = chat_completion.choices[0].message.content
    
    if debug_mode is True: 
        print(final_response)

    yes = r"\[\[Yes]]"
    no = r"\[\[No]]"
    
    if re.search(yes, final_response): 
        if debug_mode is True:
            print("Returned label 1")
        return 1
    elif re.search(no, final_response):
        if debug_mode is True:
            print("Returned label 0")
        return 0
    else:
        print("Didn't extract Yes or No!")
        # print(f"Unexpected response received: {final_response}")
        return -1
    # except requests.exceptions.RequestException as e:
    #     print(f"HTTP Request failed: {e}")
    #     return -1


############################## ANTHROPIC CUSTOM MODELS (NOT OPENAI) ######################################

            
def few_shot_context_relevance_scoring_claude(system_prompt: str, query: str, document: str, model_choice, query_id: str, debug_mode: bool, request_delay: int, few_shot_examples=None):
    ANTHROPIC_API_KEY = os.getenv("ANTHROPIC_API_KEY")

    if not ANTHROPIC_API_KEY:
        sys.exit("ANTHROPIC_API_KEY environment variable not set. Please set the variable.")

    user_prompt = ""
    if few_shot_examples is not None:
        for row in range(len(few_shot_examples)):
                    user_prompt += "Question: " + few_shot_examples.iloc[row][query_id] + "\n"
                    user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
                    current_label = few_shot_examples.iloc[row]['Context_Relevance_Label']
                    if current_label == 1 or current_label == 1.0:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label == 0 or current_label == 0.0:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"

    user_prompt += "Question: " + query + "\n"
    user_prompt += "Document: " + document + "\n"
    user_prompt += "Label: "

    time.sleep(request_delay)

    client = anthropic.Anthropic(
    api_key=ANTHROPIC_API_KEY,
    )

    chat_completion = client.messages.create(
        model=model_choice,
        max_tokens=1024,
        system=system_prompt.strip(),
        messages=[
            {
            "role": "user",
            "content": user_prompt.strip(),
            }, 
        ],
    )
    if chat_completion.content: # Check if list empty
        final_response = chat_completion.content[0].text
        if debug_mode is True: 
            print(final_response)

    else:
        return -1

    yes = r"\[\[Yes]]"
    no = r"\[\[No]]"

    if re.search(yes, final_response): 
        if debug_mode is True:
            print("Returned label 1")
        return 1
    elif re.search(no, final_response):
        if debug_mode is True:
            print("Returned label 0")
        return 0
    else:
        print("Didn't extract Yes or No!")
        # print(f"Unexpected response received: {final_response}")
        return -1

def few_shot_answer_faithfulness_scoring_claude(system_prompt: str, query: str, document: str, answer: str, model_choice, query_id: str, debug_mode: bool, request_delay: int, few_shot_examples=None):
    ANTHROPIC_API_KEY = os.getenv("ANTHROPIC_API_KEY")

    if not ANTHROPIC_API_KEY:
        sys.exit("ANTHROPIC_API_KEY environment variable not set. Please set the variable.")

    user_prompt = ""
    if few_shot_examples is not None:
        for row in range(len(few_shot_examples)):
                user_prompt += "Question: " + few_shot_examples.iloc[row][query_id] + "\n"
                user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
                user_prompt += "Answer: " + few_shot_examples.iloc[row]['Answer'] + "\n"
                current_label = few_shot_examples.iloc[row]['Answer_Faithfulness_Label']
                if current_label == 1 or current_label == 1.0:
                    warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                    current_label = "[[Yes]]"
                elif current_label == 0 or current_label == 0.0:
                    warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                    current_label = "[[No]]"
                user_prompt += "Label: " + str(current_label) + "\n\n"

    user_prompt += "Question: " + query + "\n"
    user_prompt += "Document: " + document + "\n"
    user_prompt += "Answer: " + str(answer) + "\n"
    user_prompt += "Label: "

    time.sleep(request_delay)

    client = anthropic.Anthropic(
    api_key=ANTHROPIC_API_KEY,
    )

    chat_completion = client.messages.create(
        model=model_choice,
        max_tokens=1024,
        system=system_prompt.strip(),
        messages=[
            {
            "role": "user",
            "content": user_prompt.strip(),
            }, 
        ],
    )
    if chat_completion.content:  # This checks if the list is not empty
        final_response = chat_completion.content[0].text
        if debug_mode is True: 
            print(final_response)
    else:
        return -1

    yes = r"\[\[Yes]]"
    no = r"\[\[No]]"

    
    if re.search(yes, final_response): 
        if debug_mode is True:
            print("Returned label 1")
        return 1
    elif re.search(no, final_response):
        if debug_mode is True:
            print("Returned label 0")
        return 0
    else:
        print("Didn't extract Yes or No!")
        # print(f"Unexpected response received: {final_response}")
        return -1

def few_shot_answer_relevance_scoring_claude(system_prompt: str, query: str, document: str, answer: str, model_choice, query_id: str, debug_mode: bool, request_delay: int, few_shot_examples=None):
    ANTHROPIC_API_KEY = os.getenv("ANTHROPIC_API_KEY")

    if not ANTHROPIC_API_KEY:
        sys.exit("ANTHROPIC_API_KEY environment variable not set. Please set the variable.")

    user_prompt = ""
    if few_shot_examples is not None:
        for row in range(len(few_shot_examples)):
            user_prompt += "Question: " + few_shot_examples.iloc[row][query_id] + "\n"
            user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
            user_prompt += "Answer: " + few_shot_examples.iloc[row]['Answer'] + "\n"
            current_label = few_shot_examples.iloc[row]['Answer_Relevance_Label']
            if current_label == 1 or current_label == 1.0:
                warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                current_label = "[[Yes]]"
            elif current_label == 0 or current_label == 0.0:
                warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                current_label = "[[No]]"
            user_prompt += "Label: " + str(current_label) + "\n\n"

    user_prompt += "Question: " + query + "\n"
    user_prompt += "Document: " + document + "\n"
    user_prompt += "Answer: " + str(answer) + "\n"
    user_prompt += "Label: "

    time.sleep(request_delay)

    client = anthropic.Anthropic(
    api_key=ANTHROPIC_API_KEY,
    )

    chat_completion = client.messages.create(
        model=model_choice,
        max_tokens=1024,
        system=system_prompt.strip(),
        messages=[
            {
            "role": "user",
            "content": user_prompt.strip(),
            }, 
        ],
    )
    if chat_completion.content:  # This checks if the list is not empty
        final_response = chat_completion.content[0].text
        if debug_mode is True: 
            print(final_response)
    else:
        return -1

    yes = r"\[\[Yes]]"
    no = r"\[\[No]]"

    
    if re.search(yes, final_response): 
        if debug_mode is True:
            print("Returned label 1")
        return 1
    elif re.search(no, final_response):
        if debug_mode is True:
            print("Returned label 0")
        return 0
    else:
        print("Didn't extract Yes or No!")
        # print(f"Unexpected response received: {final_response}")
        return -1


def few_shot_context_relevance_scoring_vllm(system_prompt: str, query: str, document: str, model_choice: str, query_id: str, debug_mode: bool, host_url: str, request_delay: int, few_shot_examples=None):
    for _ in range(5):
        #try:

            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += "Question: " + few_shot_examples.iloc[row][query_id] + "\n"
                    user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
                    current_label = few_shot_examples.iloc[row]['Context_Relevance_Label']
                    if current_label == 1 or current_label == 1.0:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label == 0 or current_label == 0.0:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"
            
            user_prompt += "Question: " + query + "\n"
            user_prompt += "Document: " + document + "\n"
            user_prompt += "Label: "

    time.sleep(request_delay) # Model query delay chosen by user

    openai_api_key = "EMPTY"
    client = OpenAI(
    api_key=openai_api_key,
    base_url=host_url
    )

    chat_completion = client.chat.completions.create(
    messages=[
        {
        "role": "system",
        "content": system_prompt,
        },
        {
        "role": "user",
        "content": user_prompt,
        }
    ],
    model=model_choice
    )
    final_response = chat_completion.choices[0].message.content
    
    if debug_mode is True: 
        print(final_response)

    yes = r"\[\[Yes]]"
    no = r"\[\[No]]"
    
    if re.search(yes, final_response): 
        if debug_mode is True:
            print("Returned label 1")
        return 1
    elif re.search(no, final_response):
        if debug_mode is True:
            print("Returned label 0")
        return 0
    else:
        print("Didn't extract Yes or No!")
        # print(f"Unexpected response received: {final_response}")
        return -1
    
####################################################################

def few_shot_answer_faithfulness_scoring_vllm(system_prompt: str, query: str, document: str, answer: str, model_choice: str, query_id: str, debug_mode: bool, host_url: str, request_delay: int, few_shot_examples=None):
    user_prompt = ""
    if few_shot_examples is not None:
        for row in range(len(few_shot_examples)):
            user_prompt += "Question: " + few_shot_examples.iloc[row][query_id] + "\n"
            user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
            user_prompt += "Answer: " + few_shot_examples.iloc[row]['Answer'] + "\n"
            current_label = few_shot_examples.iloc[row]['Answer_Relevance_Label']
            if current_label == 1 or current_label == 1.0:
                warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                current_label = "[[Yes]]"
            elif current_label == 0 or current_label == 0.0:
                warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                current_label = "[[No]]"
            user_prompt += "Label: " + str(current_label) + "\n\n"

    user_prompt += "Question: " + query + "\n"
    user_prompt += "Document: " + document + "\n"
    user_prompt += "Answer: " + str(answer) + "\n"
    user_prompt += "Label: "

    time.sleep(request_delay) # Model query delay chosen by user

    openai_api_key = "EMPTY"
    client = OpenAI(
    api_key=openai_api_key,
    base_url=host_url
    )

    chat_completion = client.chat.completions.create(
    messages=[
        {
        "role": "system",
        "content": system_prompt,
        },
        {
        "role": "user",
        "content": user_prompt,
        }
    ],
    model=model_choice
    )
    final_response = chat_completion.choices[0].message.content
    
    if debug_mode is True: 
        print(final_response)

    yes = r"\[\[Yes]]"
    no = r"\[\[No]]"
    
    if re.search(yes, final_response): 
        if debug_mode is True:
            print("Returned label 1")
        return 1
    elif re.search(no, final_response):
        if debug_mode is True:
            print("Returned label 0")
        return 0
    else:
        print("Didn't extract Yes or No!")
        # print(f"Unexpected response received: {final_response}")
        return -1
    
####################################################################

def few_shot_answer_relevance_scoring_vllm(system_prompt: str, query: str, document: str, answer: str, model_choice: str, query_id: str, debug_mode: bool, host_url: str, request_delay: int, few_shot_examples=None):
    user_prompt = ""
    if few_shot_examples is not None:
        for row in range(len(few_shot_examples)):
            user_prompt += "Question: " + few_shot_examples.iloc[row][query_id] + "\n"
            user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
            user_prompt += "Answer: " + few_shot_examples.iloc[row]['Answer'] + "\n"
            current_label = few_shot_examples.iloc[row]['Answer_Relevance_Label']
            if current_label == 1 or current_label == 1.0:
                warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                current_label = "[[Yes]]"
            elif current_label == 0 or current_label == 0.0:
                warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                current_label = "[[No]]"
            user_prompt += "Label: " + str(current_label) + "\n\n"

    user_prompt += "Question: " + query + "\n"
    user_prompt += "Document: " + document + "\n"
    user_prompt += "Answer: " + str(answer) + "\n"
    user_prompt += "Label: "

    time.sleep(request_delay) # Model query delay chosen by user

    openai_api_key = "EMPTY"
    client = OpenAI(
    api_key=openai_api_key,
    base_url=host_url
    )

    chat_completion = client.chat.completions.create(
    messages=[
        {
        "role": "system",
        "content": system_prompt,
        },
        {
        "role": "user",
        "content": user_prompt,
        }
    ],
    model=model_choice
    )
    final_response = chat_completion.choices[0].message.content
    
    if debug_mode is True: 
        print(final_response)

    yes = r"\[\[Yes]]"
    no = r"\[\[No]]"
    
    if re.search(yes, final_response): 
        if debug_mode is True:
            print("Returned label 1")
        return 1
    elif re.search(no, final_response):
        if debug_mode is True:
            print("Returned label 0")
        return 0
    else:
        print("Didn't extract Yes or No!")
        # print(f"Unexpected response received: {final_response}")
        return -1